# mypy: disable-error-code="import, override"
"""Defines modules for RWKV blocks.

RWKV blocks are similar to Transformer blocks, but use a different attention
mechanism that doesn't require a linearly growing KV cache.

Training requires CUDA kernel requires installing ``triton``:

.. code-block:: bash

    pip install triton
"""

import math
import os
import warnings
from typing import Callable, Literal, cast, get_args

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.autograd.function import Function, FunctionCtx, once_differentiable

from mlfab.nn.architectures.next_token import SamplingStrategy, sample_from_logits
from mlfab.nn.triton import supports_triton

WkvFnKey = Literal["eps", "log"]

RwkvAttentionState = tuple[Tensor, Tensor]
RwkvFeedForwardState = Tensor
RwkvState = tuple[RwkvAttentionState, RwkvFeedForwardState]

EPS = 1e-4


@torch.jit.script
def wkv_with_eps_forward(w: Tensor, u: Tensor, k: Tensor, v: Tensor, state: Tensor) -> tuple[Tensor, Tensor]:
    bsz, tsz, chans = k.shape

    assert w.shape == u.shape == (chans,)
    assert v.shape == (bsz, tsz, chans)
    assert state.shape == (bsz, 3, 1, chans)

    alpha, beta, eps = state[:, :, -1].chunk(3, dim=1)  # (B, 1, D), (B, 1, D), (B, 1, D)

    _, tsz, _ = k.shape

    wkvs = []
    alphas = [alpha]
    betas = [beta]
    epss = [eps]

    for t in range(tsz):
        kt, vt = k[:, t : t + 1], v[:, t : t + 1]
        ukt = u + kt
        tau = torch.maximum(ukt, eps)
        e1 = torch.exp(eps - tau)
        e2 = torch.exp(ukt - tau)
        wkv = (e1 * alpha + e2 * vt) / (e1 * beta + e2)
        wkvs.append(wkv)

        w_eps = eps - w
        eps = torch.maximum(w_eps, kt)
        e1 = torch.exp(w_eps - eps)
        e2 = torch.exp(kt - eps)
        alpha = e1 * alpha + e2 * vt
        beta = e1 * beta + e2

        alphas.append(alpha)
        betas.append(beta)
        epss.append(eps)

    alpha = torch.stack(alphas, dim=2)
    beta = torch.stack(betas, dim=2)
    eps = torch.stack(epss, dim=2)

    return torch.cat(wkvs, 1), torch.cat((alpha, beta, eps), dim=1)


@torch.jit.script
def wkv_with_eps_backward(
    w: Tensor,
    u: Tensor,
    k: Tensor,
    v: Tensor,
    state: Tensor,
    grad_wkv: Tensor,
    grad_state: Tensor,
) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
    bsz, tsz, chans = k.shape

    assert w.shape == u.shape == (chans,)
    assert v.shape == (bsz, tsz, chans)
    assert state.shape == (bsz, 3, tsz + 1, chans)
    assert grad_wkv.shape == (bsz, tsz, chans)
    assert grad_state.shape == (bsz, 3, 1, chans)

    alpha, beta, eps = state.chunk(3, dim=1)  # (B, 1, T + 1, D), (B, 1, T + 1, D), (B, 1, T + 1, D)
    grad_alpha, grad_beta, grad_eps = grad_state[:, :, 0].chunk(3, dim=1)  # (B, 1, D), (B, 1, D), (B, 1, D)
    grad_eps = grad_eps.clone()

    grad_w = torch.zeros_like(w)
    grad_u = torch.zeros_like(u)
    grad_k = torch.zeros_like(k)
    grad_v = torch.zeros_like(v)

    for t in range(tsz - 1, -1, -1):
        kt, vt = k[:, t : t + 1], v[:, t : t + 1]
        alpha_prev, beta_prev, eps_prev = alpha[:, :, t], beta[:, :, t], eps[:, :, t]
        alpha_curr, beta_curr, eps_curr = alpha[:, :, t + 1], beta[:, :, t + 1], eps[:, :, t + 1]
        ukt = u + kt
        tau = torch.maximum(ukt, eps_prev)
        e1 = torch.exp(eps_prev - tau)
        e2 = torch.exp(ukt - tau)

        euke = torch.exp(ukt + eps_prev - 2 * tau)

        denom = e1 * beta_prev + e2
        denom_sq = denom * denom

        grad_wkvt = grad_wkv[:, t : t + 1]

        # Backpropagates wkv gradients.
        grad_uk = grad_wkvt * e2 * (e1 * beta_prev * vt - e1 * alpha_prev) / denom_sq
        grad_u += grad_uk.flatten(0, -2).sum(0)
        grad_k[:, t : t + 1] += grad_uk
        grad_v[:, t : t + 1] += grad_wkvt * e2 / denom

        grad_alpha_wkv = grad_wkvt * e1 / denom
        grad_beta_wkv = -grad_wkvt * e1 * (e2 * vt + e1 * alpha_prev) / denom_sq
        grad_eps_wkv = grad_wkvt * euke * (alpha_prev - vt * beta_prev) / (e1 * beta_prev + e2) ** 2

        e1 = torch.exp(eps_prev - eps_curr - w)
        e2 = torch.exp(kt - eps_curr)

        # Backpropagates alpha gradients.
        grad_alpha_we = grad_alpha * e1 * alpha_prev
        grad_w -= grad_alpha_we.flatten(0, -2).sum(0)
        grad_k[:, t : t + 1] += grad_alpha * e2 * vt
        grad_v[:, t : t + 1] += grad_alpha * e2
        grad_eps += grad_alpha * -alpha_curr

        # Backpropagates beta gradients.
        grad_beta_we = grad_beta * e1 * beta_prev
        grad_w -= grad_beta_we.flatten(0, -2).sum(0)
        grad_k[:, t : t + 1] += grad_beta * e2
        grad_eps += grad_beta * -beta_curr

        # Backpropagates epsilon gradients.
        eps_grad_mask = eps_prev - w > kt
        grad_eps_we = torch.where(eps_grad_mask, grad_eps, torch.zeros_like(grad_eps))
        grad_w -= grad_eps_we.flatten(0, -2).sum(0)
        grad_k[:, t : t + 1] += torch.where(eps_grad_mask, torch.zeros_like(grad_eps), grad_eps)

        # Computes gradients for alpha, beta and epsilon.
        grad_alpha = grad_alpha * e1 + grad_alpha_wkv
        grad_beta = grad_beta * e1 + grad_beta_wkv
        grad_eps = grad_alpha_we + grad_beta_we + grad_eps_we + grad_eps_wkv

    return grad_w, grad_u, grad_k, grad_v, torch.stack((grad_alpha, grad_beta, grad_eps), dim=1)


class WkvWithEps(Function):
    @staticmethod
    def forward(
        ctx: FunctionCtx,
        w: Tensor,
        u: Tensor,
        k: Tensor,
        v: Tensor,
        state: Tensor,
    ) -> tuple[Tensor, Tensor]:
        wkv, state_out = wkv_with_eps_forward(w, u, k, v, state)
        ctx.save_for_backward(w, u, k, v, state_out)
        return wkv, state_out[:, :, -1:]

    @staticmethod
    @once_differentiable
    def backward(
        ctx: FunctionCtx,
        grad_wkv: Tensor,
        grad_state: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        w, u, k, v, state = cast(tuple[Tensor, ...], ctx.saved_tensors)
        return wkv_with_eps_backward(w, u, k, v, state, grad_wkv, grad_state)


def initial_state_with_eps(emb_dim: int) -> Tensor:
    return torch.zeros(1, 3, 1, emb_dim)


def wkv_with_eps(w: Tensor, u: Tensor, k: Tensor, v: Tensor, state: Tensor) -> tuple[Tensor, Tensor]:
    """Runs the core WKV computation.

    Args:
        w: The decay tensor, with shape (D)
        u: The output multiplier tensor, with shape (D)
        k: The K tensor, with shape (B, T, D)
        v: The V tensor, with shape (B, T, D)
        state: The state tensor, with shape (B, 3, T, D), consisting of the
            alpha, beta and eps tensors, each with shape (B, 1, T, D)

    Returns:
        The WKV tensor, with shape (B, T, D), and the next state, with shape
        (B, 3, 1, D), consisting of the next alpha, beta and eps tensors, each
        with shape (B, 1, 1, D)
    """
    return WkvWithEps.apply(w, u, k, v, state)


@torch.jit.script
def logaddexp(a: Tensor, b: Tensor) -> Tensor:
    max_ab = torch.maximum(a, b)
    return max_ab + torch.log(torch.exp(a - max_ab) + torch.exp(b - max_ab))


@torch.jit.script
def logsubexp(a: Tensor, b: Tensor, log_eps: float) -> Tensor:
    max_ab = torch.clamp_min(torch.maximum(a, b), log_eps)
    return max_ab + torch.log(torch.exp(a - max_ab) - torch.exp(b - max_ab))


@torch.jit.script
def wkv_log_space_forward(
    w: Tensor,
    u: Tensor,
    k: Tensor,
    v: Tensor,
    state: Tensor,
    eps: float = EPS,
    normalize: bool = False,
) -> tuple[Tensor, Tensor]:
    bsz, tsz, chans = k.shape

    assert w.shape == u.shape == (chans,)
    assert v.shape == (bsz, tsz, chans)
    assert state.shape == (bsz, 3, 1, chans)

    ln_alpha_p, ln_alpha_m, ln_beta = state[:, :, -1].chunk(3, dim=1)

    log_eps = math.log(eps)

    wkvs = []
    ln_alpha_ps = [ln_alpha_p]
    ln_alpha_ms = [ln_alpha_m]
    ln_betas = [ln_beta]

    for t in range(tsz):
        kt, vt = k[:, t : t + 1], v[:, t : t + 1]
        vt_p, vt_m = torch.clamp_min(vt, 0) + eps, torch.clamp_min(-vt, 0) + eps
        ln_v_p, ln_v_m = torch.log(vt_p), torch.log(vt_m)

        if normalize:
            ln_alpha_pm = torch.minimum(ln_alpha_p, ln_alpha_m) - eps
            ln_alpha_p = logsubexp(ln_alpha_p, ln_alpha_pm, log_eps)
            ln_alpha_m = logsubexp(ln_alpha_m, ln_alpha_pm, log_eps)

        ln_wkv_p = logaddexp(u + kt + ln_v_p, ln_alpha_p) - logaddexp(u + kt, ln_beta)
        ln_wkv_m = logaddexp(u + kt + ln_v_m, ln_alpha_m) - logaddexp(u + kt, ln_beta)

        wkv = torch.exp(ln_wkv_p) - torch.exp(ln_wkv_m)
        wkvs.append(wkv)

        ln_alpha_p = logaddexp(ln_alpha_p - w, kt + ln_v_p)
        ln_alpha_m = logaddexp(ln_alpha_m - w, kt + ln_v_m)
        ln_beta = logaddexp(ln_beta - w, kt)

        ln_alpha_ps.append(ln_alpha_p)
        ln_alpha_ms.append(ln_alpha_m)
        ln_betas.append(ln_beta)

    ln_alpha_p = torch.stack(ln_alpha_ps, dim=2)
    ln_alpha_m = torch.stack(ln_alpha_ms, dim=2)
    ln_beta = torch.stack(ln_betas, dim=2)

    return torch.cat(wkvs, 1), torch.cat((ln_alpha_p, ln_alpha_m, ln_beta), dim=1)


@torch.jit.script
def wkv_log_space_backward(
    w: Tensor,
    u: Tensor,
    k: Tensor,
    v: Tensor,
    state: Tensor,
    grad_wkv: Tensor,
    grad_state: Tensor,
    eps: float = EPS,
) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
    bsz, tsz, chans = k.shape

    assert w.shape == u.shape == (chans,)
    assert v.shape == (bsz, tsz, chans)
    assert state.shape == (bsz, 3, tsz, chans)
    assert grad_wkv.shape == (bsz, tsz, chans)
    assert grad_state.shape == (bsz, 3, 1, chans)

    grad_ln_alpha_p, grad_ln_alpha_m, grad_ln_beta = grad_state[:, :, 0].chunk(3, dim=1)

    grad_w = torch.zeros_like(w)
    grad_u = torch.zeros_like(u)
    grad_k = torch.zeros_like(k)
    grad_v = torch.zeros_like(v)

    for t in range(tsz - 1, -1, -1):
        kt, vt = k[:, t : t + 1], v[:, t : t + 1]
        vt_p, vt_m = torch.clamp_min(vt, 0) + eps, torch.clamp_min(-vt, 0) + eps
        ln_v_p, ln_v_m = torch.log(vt_p), torch.log(vt_m)

        ln_alpha_p_prev, ln_alpha_m_prev, ln_beta_prev = state[:, :, t].chunk(3, dim=1)

        uk = u + kt
        ukv_p, ukv_m = uk + ln_v_p, uk + ln_v_m

        ukb = logaddexp(uk, ln_beta_prev)
        wkv_p = torch.exp(logaddexp(ukv_p, ln_alpha_p_prev) - ukb)
        wkv_m = torch.exp(logaddexp(ukv_m, ln_alpha_m_prev) - ukb)

        grad_wkvt = grad_wkv[:, t : t + 1]
        grad_ln_wkv_p, grad_ln_wkv_m = grad_wkvt * wkv_p, grad_wkvt * -wkv_m

        # Backpropagates wkv gradients.
        e_num_p = torch.exp(ln_alpha_p_prev - ukv_p)
        e_num_m = torch.exp(ln_alpha_m_prev - ukv_m)
        e_den = torch.exp(ln_beta_prev - uk)
        grad_wkv_den_p = grad_ln_wkv_p / (1 + e_den)
        grad_wkv_den_m = grad_ln_wkv_m / (1 + e_den)
        grad_kv_p = grad_ln_wkv_p / (1 + e_num_p)
        grad_kv_m = grad_ln_wkv_m / (1 + e_num_m)
        grad_uk = grad_kv_p + grad_kv_m - grad_wkv_den_p - grad_wkv_den_m
        grad_u += grad_uk.flatten(0, -2).sum(0)
        grad_k[:, t : t + 1] += grad_uk
        grad_v[:, t : t + 1] += torch.where(vt > 0, grad_kv_p / vt_p, grad_kv_m / -vt_m)

        grad_ln_alpha_wkv_p = grad_ln_wkv_p / (1 + (1 / e_num_p))
        grad_ln_alpha_wkv_m = grad_ln_wkv_m / (1 + (1 / e_num_m))
        grad_ln_beta_wkv = -grad_ln_wkv_p / (1 + (1 / e_den)) - grad_ln_wkv_m / (1 + (1 / e_den))

        # Backpropagates alpha gradients.
        e_alpha_p = torch.exp(kt + ln_v_p + w - ln_alpha_p_prev)
        e_alpha_m = torch.exp(kt + ln_v_m + w - ln_alpha_m_prev)
        grad_wa_p = grad_ln_alpha_p / (1 + e_alpha_p)
        grad_wa_m = grad_ln_alpha_m / (1 + e_alpha_m)
        grad_w -= (grad_wa_p + grad_wa_m).flatten(0, -2).sum(0)
        grad_kv_p = grad_ln_alpha_p / (1 + (1 / e_alpha_p))
        grad_kv_m = grad_ln_alpha_m / (1 + (1 / e_alpha_m))
        grad_k[:, t : t + 1] += grad_kv_p + grad_kv_m
        grad_v[:, t : t + 1] += torch.where(vt > 0, grad_kv_p / vt_p, -grad_kv_m / vt_m)

        # Backpropagates beta gradients.
        e_beta = torch.exp(kt + w - ln_beta_prev)
        grad_wb = grad_ln_beta / (1 + e_beta)
        grad_w -= grad_wb.flatten(0, -2).sum(0)
        grad_k[:, t : t + 1] += grad_ln_beta / (1 + (1 / e_beta))

        # Compute gradients for log alpha and log beta.
        grad_ln_alpha_p = grad_wa_p + grad_ln_alpha_wkv_p
        grad_ln_alpha_m = grad_wa_m + grad_ln_alpha_wkv_m
        grad_ln_beta = grad_wb + grad_ln_beta_wkv

    return grad_w, grad_u, grad_k, grad_v, torch.stack((grad_ln_alpha_p, grad_ln_alpha_m, grad_ln_beta), dim=1)


class WkvLogSpace(Function):
    @staticmethod
    def forward(
        ctx: FunctionCtx,
        w: Tensor,
        u: Tensor,
        k: Tensor,
        v: Tensor,
        state: Tensor,
    ) -> tuple[Tensor, Tensor]:
        wkv, state_out = wkv_log_space_forward(w, u, k, v, state)
        ctx.save_for_backward(w, u, k, v, state_out[:, :, :-1])
        return wkv, state_out[:, :, -1:]

    @staticmethod
    @once_differentiable
    def backward(
        ctx: FunctionCtx,
        grad_wkv: Tensor,
        grad_state: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        w, u, k, v, state = cast(tuple[Tensor, ...], ctx.saved_tensors)
        return wkv_log_space_backward(w, u, k, v, state, grad_wkv, grad_state)


def initial_state_log_space(emb_dim: int) -> Tensor:
    return torch.full((1, 3, 1, emb_dim), float("-inf"))


def wkv_log_space(w: Tensor, u: Tensor, k: Tensor, v: Tensor, state: Tensor) -> tuple[Tensor, Tensor]:
    """Runs the core WKV computation.

    Args:
        w: The decay tensor, with shape (D)
        u: The output multiplier tensor, with shape (D)
        k: The K tensor, with shape (B, T, D)
        v: The V tensor, with shape (B, T, D)
        state: The state tensor, with shape (B, 3, D), consisting of the
            alpha plus, alpha minus and beta tensors, each with shape (B, 1, D)

    Returns:
        The WKV tensor, with shape (B, T, D), and the next state, with shape
        (B, 2, D), consisting of the next alpha plus, alpha minus and beta
        tensors, each with shape (B, 1, D)
    """
    return WkvLogSpace.apply(w, u, k, v, state)


def get_wkv_fn(key: WkvFnKey) -> Callable[[Tensor, Tensor, Tensor, Tensor, Tensor], tuple[Tensor, Tensor]]:
    match key:
        case "eps":
            return wkv_with_eps
        case "log":
            return wkv_log_space
        case _:
            raise ValueError(f"Unsupported key: {key}")


def get_wkv_fn_cuda(key: WkvFnKey) -> Callable[[Tensor, Tensor, Tensor, Tensor, Tensor], tuple[Tensor, Tensor]]:
    if not supports_triton():
        return get_wkv_fn(key)

    from mlfab.nn.triton.rwkv import wkv_triton_log_space, wkv_triton_with_eps

    match key:
        case "eps":
            return wkv_triton_with_eps
        case "log":
            return wkv_triton_log_space
        case _:
            raise ValueError(f"Unsupported key: {key}")


def get_default_wkv_fn_key() -> WkvFnKey:
    if "WKV_FN" in os.environ:
        assert (wkv_fn_str := os.environ["WKV_FN"]) in get_args(WkvFnKey), f"Unsupported WKV_FN: {wkv_fn_str}"
        return cast(WkvFnKey, wkv_fn_str)

    warnings.warn("Using default WKV_FN: eps")
    return "eps"


class RwkvAttention(nn.Module):
    init_x: Tensor
    init_state: Tensor

    def __init__(self, dim: int, wkv_key: WkvFnKey | None = None) -> None:
        super().__init__()

        self.time_decay = nn.Parameter(torch.ones(dim))
        self.time_first = nn.Parameter(torch.ones(dim))

        self.time_mix_k = nn.Parameter(torch.ones(1, 1, dim))
        self.time_mix_v = nn.Parameter(torch.ones(1, 1, dim))
        self.time_mix_r = nn.Parameter(torch.ones(1, 1, dim))

        self.key = nn.Linear(dim, dim, False)
        self.value = nn.Linear(dim, dim, False)
        self.receptance = nn.Linear(dim, dim, False)
        self.output = nn.Linear(dim, dim, False)

        if wkv_key is None:
            wkv_key = get_default_wkv_fn_key()

        self.wkv_fn = get_wkv_fn(wkv_key)
        self.wkv_fn_cuda = get_wkv_fn_cuda(wkv_key)

        self.register_buffer("init_x", torch.zeros(1, 1, dim), persistent=False)
        self.register_buffer("init_state", initial_state_with_eps(dim), persistent=False)

    def reset_parameters(self) -> None:
        nn.init.orthogonal_(self.key.weight)
        nn.init.orthogonal_(self.receptance.weight)
        nn.init.orthogonal_(self.value.weight)
        nn.init.orthogonal_(self.output.weight)

    def time_shift(self, last_x: Tensor, x: Tensor) -> Tensor:
        _, tsz, _ = x.shape
        if tsz > 1:
            last_x = torch.cat((last_x, x[..., :-1, :]), dim=-2)
        return last_x

    def forward(self, x: Tensor, state: RwkvAttentionState | None) -> tuple[Tensor, RwkvAttentionState]:
        bsz, _, _ = x.shape

        if state is None:
            last_x = self.init_x.repeat_interleave(bsz, dim=0)
            last_state = self.init_state.repeat_interleave(bsz, dim=0)
        else:
            last_x, last_state = state
        last_x = self.time_shift(last_x, x)

        k = self.key(x * self.time_mix_k + last_x * (1 - self.time_mix_k))
        v = self.value(x * self.time_mix_v + last_x * (1 - self.time_mix_v))
        r = self.receptance(x * self.time_mix_r + last_x * (1 - self.time_mix_r))
        sr = torch.sigmoid(r)

        w, u = self.time_decay, self.time_first
        w = torch.exp(w)
        wkv_fn = self.wkv_fn_cuda if x.is_cuda else self.wkv_fn
        wkv, next_state = wkv_fn(w, u, k, v, last_state)
        rwkv = wkv * sr

        return self.output(rwkv), (x[..., -1:, :], next_state)


class RwkvFeedForward(nn.Module):
    init_state: Tensor

    def __init__(self, dim: int, ffn_dim: int) -> None:
        super().__init__()

        self.time_mix_k = nn.Parameter(torch.ones(1, 1, dim))
        self.time_mix_r = nn.Parameter(torch.ones(1, 1, dim))

        self.key = nn.Linear(dim, ffn_dim, False)
        self.receptance = nn.Linear(dim, dim, False)
        self.value = nn.Linear(ffn_dim, dim, False)

        self.register_buffer("init_state", torch.zeros(1, 1, dim), persistent=False)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.orthogonal_(self.key.weight)
        nn.init.orthogonal_(self.receptance.weight)
        nn.init.orthogonal_(self.value.weight)

    def time_shift(self, last_x: Tensor, x: Tensor) -> Tensor:
        _, tsz, _ = x.shape
        if tsz > 1:
            last_x = torch.cat((last_x, x[..., :-1, :]), dim=-2)
        return last_x

    def forward(self, x: Tensor, state: RwkvFeedForwardState | None = None) -> tuple[Tensor, RwkvFeedForwardState]:
        bsz = x.shape[0]

        last_x = self.time_shift(self.init_state.repeat(bsz, 1, 1) if state is None else state, x)

        k = self.key(x * self.time_mix_k + last_x * (1 - self.time_mix_k))
        r = self.receptance(x * self.time_mix_r + last_x * (1 - self.time_mix_r))
        vk = self.value(F.relu(k) ** 2)

        return torch.sigmoid(r) * vk, x[..., -1:, :]


class RwkvBlock(nn.Module):
    def __init__(
        self,
        emb_dim: int,
        pre_norm: bool = False,
        wkv_key: WkvFnKey | None = None,
        feedforward_factor: int = 4,
    ) -> None:
        super().__init__()

        self.ln0 = nn.LayerNorm(emb_dim) if pre_norm else None
        self.ln1 = nn.LayerNorm(emb_dim)
        self.ln2 = nn.LayerNorm(emb_dim)
        self.att = RwkvAttention(emb_dim, wkv_key=wkv_key)
        self.ffn = RwkvFeedForward(emb_dim, emb_dim * feedforward_factor)

    def run_attn(self, x: Tensor, state: RwkvState | None = None) -> tuple[Tensor, RwkvAttentionState]:
        return self.att.forward(self.ln1(x), None if state is None else state[0])

    def run_ffn(self, x: Tensor, state: RwkvState | None = None) -> tuple[Tensor, RwkvFeedForwardState]:
        return self.ffn.forward(self.ln2(x), None if state is None else state[1])

    def forward(self, x: Tensor, state: RwkvState | None = None) -> tuple[Tensor, RwkvState]:
        if self.ln0 is not None:
            x = self.ln0(x)
        dx, att_state_out = self.run_attn(x, state)
        x = x + dx
        dx, ffn_state_out = self.run_ffn(x, state)
        x = x + dx
        return x, (att_state_out, ffn_state_out)


class RwkvStack(nn.Module):
    """Defines a stack of RWKV modules.

    Parameters:
        emb_dim: The number of embedding dimensions in each block
        num_layers: The number of layers in the stack
        wkv_key: The WKV algorithm to use
        feedforward_factor: The factor by which the input number of dimensions
            is multiplied to get the feedforward hidden dimension.

    Inputs:
        x: The input tensor, with shape ``(B, T, D)``
        state: The previous state

    Outputs:
        The output tensor, with shape ``(B, T, D)``, and the next state
    """

    def __init__(
        self,
        emb_dim: int,
        num_layers: int,
        wkv_key: WkvFnKey | None = None,
        feedforward_factor: int = 4,
    ) -> None:
        super().__init__()

        self.blocks = nn.ModuleList(
            [
                RwkvBlock(
                    emb_dim,
                    pre_norm=i == 0,
                    wkv_key=wkv_key,
                    feedforward_factor=feedforward_factor,
                )
                for i in range(num_layers)
            ]
        )

    def forward(self, x: Tensor, state: list[RwkvState] | None = None) -> tuple[Tensor, list[RwkvState]]:
        state_out: list[RwkvState] = []
        for i, block in enumerate(self.blocks):
            x, state_out_i = block(x, None if state is None else state[i])
            state_out.append(state_out_i)
        return x, state_out


class NextTokenRwkv(nn.Module):
    """Defines a next token prediction RWKV module.

    This seems to be the most popular architecture for solving a large number
    of problems. This provides a tested implementation of the next token
    prediction transformer.
    """

    def __init__(
        self,
        emb_dim: int,
        num_layers: int,
        vocab_size: int,
        wkv_key: WkvFnKey | None = None,
        feedforward_factor: int = 4,
    ) -> None:
        super().__init__()

        self.init_emb = nn.Parameter(torch.randn(1, 1, emb_dim) * 0.02)
        self.embeddings = nn.Embedding(vocab_size, emb_dim)
        self.rwkv = RwkvStack(
            emb_dim=emb_dim,
            num_layers=num_layers,
            wkv_key=wkv_key,
            feedforward_factor=feedforward_factor,
        )
        self.proj = nn.Linear(emb_dim, vocab_size)

    def forward(self, tokens_bt: Tensor) -> Tensor:
        x_btc = self.embeddings(tokens_bt[:, :-1])
        x_btc = torch.cat((self.init_emb.expand(x_btc.size(0), 1, -1), x_btc), dim=1)
        x_btc, _ = self.rwkv(x_btc)
        logits_btc = self.proj(x_btc)
        return logits_btc

    def infer(
        self,
        t: int,
        bsz: int = 1,
        sampling_strategy: SamplingStrategy = "top-p",
        k: int | None = None,
        p: float | None = 0.95,
        temperature: float = 1.0,
    ) -> Tensor:
        x_b1c: Tensor = self.init_emb.expand(bsz, 1, -1)
        x_b1c, state = self.rwkv(x_b1c)
        logits_b1l = self.proj(x_b1c)
        tokens_bt = sample_from_logits(logits_b1l, sampling_strategy, k=k, p=p, temperature=temperature)
        tokens_b1 = tokens_bt[:, :1]

        for _ in range(1, t):
            x_b1c = self.embeddings(tokens_b1)
            x_b1c, state = self.rwkv(x_b1c, state)
            logits_b1l = self.proj(x_b1c)
            tokens_b1 = sample_from_logits(logits_b1l, sampling_strategy, k=k, p=p, temperature=temperature)
            tokens_bt = torch.cat((tokens_bt, tokens_b1), dim=1)

        return tokens_bt


class NextTokenWithEmbeddingsRwkv(nn.Module):
    """Defines a next token prediction RWKV module, over base embeddings.

    This is similar to the ``NextTokenTransformer`` except that each of the
    input timesteps also has an associated embedding, which is added to the
    input before the RWKV layers.
    """

    def __init__(
        self,
        emb_dim: int,
        num_layers: int,
        vocab_size: int,
        wkv_key: WkvFnKey | None = None,
        feedforward_factor: int = 4,
    ) -> None:
        super().__init__()

        self.init_emb = nn.Parameter(torch.randn(1, 1, emb_dim) * 0.02)
        self.embeddings = nn.Embedding(vocab_size, emb_dim)
        self.rwkv = RwkvStack(
            emb_dim=emb_dim,
            num_layers=num_layers,
            wkv_key=wkv_key,
            feedforward_factor=feedforward_factor,
        )
        self.proj = nn.Linear(emb_dim, vocab_size)

    def forward(self, tokens_bt: Tensor, emb_btc: Tensor) -> Tensor:
        x_btc = self.embeddings(tokens_bt[:, :-1])
        x_btc = torch.cat((self.init_emb.expand(x_btc.size(0), 1, -1), x_btc), dim=1)
        x_btc = x_btc + emb_btc
        x_btc, _ = self.rwkv(x_btc)
        logits_btc = self.proj(x_btc)
        return logits_btc

    def infer(
        self,
        emb_btc: Tensor,
        sampling_strategy: SamplingStrategy = "top-p",
        k: int | None = None,
        p: float | None = 0.95,
        temperature: float = 1.0,
    ) -> Tensor:
        x_b1c: Tensor = self.init_emb.expand(emb_btc.size(0), 1, -1)
        x_b1c = x_b1c + emb_btc[:, :1]
        x_b1c, state = self.rwkv(x_b1c)
        logits_b1l = self.proj(x_b1c)
        tokens_bt = sample_from_logits(logits_b1l, sampling_strategy, k=k, p=p, temperature=temperature)
        tokens_b1 = tokens_bt[:, :1]

        for t in range(1, emb_btc.size(1)):
            x_b1c = self.embeddings(tokens_b1) + emb_btc[:, t : t + 1]
            x_b1c, state = self.rwkv(x_b1c, state)
            logits_b1l = self.proj(x_b1c)
            tokens_b1 = sample_from_logits(logits_b1l, sampling_strategy, k=k, p=p, temperature=temperature)
            tokens_bt = torch.cat((tokens_bt, tokens_b1), dim=1)

        return tokens_bt



"""Defines a mixin for doing FP16 scaling.

FP16 scaling is a technique for training with FP16 precision while maintaining
FP32 precision for the model weights. This is done by scaling the loss by a
large factor (e.g. 2^16) and then scaling the gradients by the inverse of that
factor. So if the scale factor starts to decrease, it means that the loss is
overflowing and training is diverging.
"""

import functools
import json
import logging
from dataclasses import dataclass
from typing import Any, ContextManager, Generic, Sequence, TypeVar

import torch
from torch import Tensor
from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
from torch.optim import Optimizer

from mlfab.core.conf import field
from mlfab.task.mixins.device import DeviceConfig, DeviceMixin
from mlfab.task.mixins.logger import LoggerConfig, LoggerMixin
from mlfab.utils.experiments import MinGradScaleError, NaNError, clip_grad_norm_, get_weight_norm

logger = logging.getLogger(__name__)


@dataclass
class GradScalerConfig:
    enabled: bool = field(True, help="If set, should FP16 training be enabled")
    init_scale: float = field(2.0**16, help="Initial scaling factor")
    growth_factor: float = field(2.0, help="Factor by which the scale is multiplied if no gradient NaNs occur")
    backoff_factor: float = field(0.5, help="Factor by which the scale is multiplied if gradient NaNs occur")
    growth_interval: int = field(2000, help="How often to grow the scale")
    min_grad_scale: float = field(1e-4, help="Minimum allowable gradient scale")
    foreach: bool | None = field(None, help="If set, use foreach implementation")


@dataclass
class MixedPrecisionConfig(DeviceConfig, LoggerConfig):
    grad_scaler: GradScalerConfig = field(GradScalerConfig(), help="Gradient scaler configuration")
    clip_grad_norm: float = field(10.0, help="What to clip the gradient norm to")
    clip_grad_norm_type: Any = field(2, help="Type of norm to use")


Config = TypeVar("Config", bound=MixedPrecisionConfig)


class MixedPrecisionMixin(DeviceMixin[Config], LoggerMixin[Config], Generic[Config]):
    """Defines a trainer mixin for doing FP16 scaling."""

    @functools.cached_property
    def grad_scaler(self) -> torch.cuda.amp.GradScaler | None:
        if self.device.supports_grad_scaler() and self.config.grad_scaler.enabled:
            return torch.cuda.amp.GradScaler(
                init_scale=self.config.grad_scaler.init_scale,
                growth_factor=self.config.grad_scaler.growth_factor,
                backoff_factor=self.config.grad_scaler.backoff_factor,
                growth_interval=self.config.grad_scaler.growth_interval,
                enabled=True,
            )

        return None

    @functools.cached_property
    def autocast_context(self) -> ContextManager:
        return self.device.autocast_context(enabled=self.config.grad_scaler.enabled)

    def scale_mixed_precision(self, tensor: Tensor) -> Tensor:
        if self.grad_scaler is not None:
            return self.grad_scaler.scale(tensor)
        return tensor

    def backward_grads(
        self,
        loss: Tensor,
        retain_graph: bool | None = None,
        inputs: Sequence[Tensor] | None = None,
    ) -> None:
        if self.grad_scaler is not None:
            loss = self.grad_scaler.scale(loss)
        if loss.numel() > 1:
            loss = loss.sum()
        isnan = not bool(torch.isfinite(loss))
        if isnan:
            loss.backward(torch.zeros_like(loss), retain_graph=retain_graph, inputs=inputs)
        else:
            loss.backward(retain_graph=retain_graph, inputs=inputs)

        if isnan:
            if any(not torch.isfinite(p).all() for p in self.parameters()):
                raise NaNError("One or more model parameters are NaN")
            if self.grad_scaler is not None:
                with torch.no_grad():
                    new_scale = self.grad_scaler.get_scale() * self.grad_scaler.get_backoff_factor()
                    if new_scale < self.config.grad_scaler.min_grad_scale:
                        raise MinGradScaleError("Minimum gradient scale reached; your loss is probably exploding")
                    logger.warning("Loss NaNs detected; reducing scale to %.2g", new_scale)
                    self.grad_scaler.update(new_scale)

    @torch.no_grad()
    def step_optimizer(self, optim: Optimizer, num_steps: int = 1) -> None:
        clip_norm = self.config.clip_grad_norm
        norm_type = self.config.clip_grad_norm_type

        # When accumulating multiple steps of gradients per backward pass, we
        # need to divide the gradients by the number of steps.
        if num_steps > 1:
            for p in self.parameters():
                if p.grad is not None:
                    p.grad /= num_steps

        # Unscale gradients.
        if self.grad_scaler is not None:
            self.grad_scaler.unscale_(optim)

        # Clips gradients.
        if isinstance(self, FSDP):
            total_norm = self.clip_grad_norm_(clip_norm, norm_type)
            was_clipped = bool(torch.isfinite(total_norm))
        else:
            total_norm, was_clipped = clip_grad_norm_(
                self.parameters(),
                max_norm=clip_norm,
                norm_type=norm_type,
                foreach=self.config.grad_scaler.foreach,
            )

        # Logs weight and gradient norms.
        self.log_scalar("weight_norm", lambda: get_weight_norm(self.parameters()), namespace="📉 optim")
        self.log_scalar("grad_norm", total_norm, namespace="📉 optim")

        # Steps the optimizer.
        if self.grad_scaler is None:
            if was_clipped:
                optim.step()
        elif was_clipped:
            self.grad_scaler.step(optim)
            self.grad_scaler.update()
        else:
            if any(not torch.isfinite(p).all() for p in self.parameters()):
                raise NaNError("One or more model parameters are NaN")
            new_scale = self.grad_scaler.get_scale() * self.grad_scaler.get_backoff_factor()
            bad = [k for k, v in self.named_parameters() if v.grad is not None and not torch.isfinite(v.grad).all()]
            bad_str = ", ".join(bad[:5])
            if len(bad) > 5:
                bad_str += f" (plus {len(bad) - 5} more)"
            bad_str += f" out of {sum(1 for _ in self.parameters())} parameters"
            if new_scale < self.config.grad_scaler.min_grad_scale:
                raise MinGradScaleError(f"Minimum gradient scale reached. Bad parameters: {bad_str}")
            logger.warning("Gradient NaNs detected for %s; reducing scale to %.2g", bad_str, new_scale)
            self.grad_scaler.update(new_scale)

    def log_mp_scale(self) -> None:
        if (scaler := self.grad_scaler) is not None and scaler._enabled:
            self.log_scalar("scale", scaler.get_scale, namespace="⚖️ fp16")
            self.log_scalar("growth", scaler._get_growth_tracker, namespace="⚖️ fp16")

    def load_task_state_dict(self, state_dict: dict, strict: bool = True, assign: bool = False) -> None:
        if self.grad_scaler is not None and "grad_scaler" in state_dict:
            self.grad_scaler.load_state_dict(json.loads(state_dict["grad_scaler"]))
        super().load_task_state_dict(state_dict, strict, assign)

    def task_state_dict(self) -> dict:
        state_dict = super().task_state_dict()
        if self.grad_scaler is not None:
            assert "grad_scaler" not in state_dict, "Duplicate keys!"
            state_dict["grad_scaler"] = json.dumps(self.grad_scaler.state_dict())
        return state_dict



"""Defines some common functions for next-token prediction."""

from typing import Literal, overload

import torch
import torch.nn.functional as F
from torch import Tensor

MaskMode = Literal["causal", "lengths", "combine"]

SamplingStrategy = Literal["top-p", "top-k", "greedy"]


def top_p_sampling(logits_btc: Tensor, p: float, temperature: float = 1.0, dim: int = -1) -> Tensor:
    """Samples from a distribution using top-P sampling.

    This is a modified version of ``torch.multinomial`` that uses top-p
    sampling instead of top-k sampling. The difference is that top-k sampling
    sets the probability of all values outside the top-k to zero, whereas
    top-p sampling sets the probability of all values outside the top-p
    to zero.

    Parameters:
        logits_btc: The input tensor, of shape ``(B, T, C)``.
        p: The probability threshold.
        temperature: The temperature to apply to the logits.
        dim: The dimension to sample from. Defaults to ``-1``.

    Returns:
        The sampled indices, of shape ``(B, T)``.
    """
    with torch.no_grad():
        assert 0.0 <= p <= 1.0, f"`{p=}` must be between 0 and 1"
        if dim != -1:
            logits_btc = logits_btc.transpose(dim, -1)
        orig_shape = logits_btc.shape[:-1]
        logits_nl = logits_btc.flatten(0, -2)
        probs_nl = F.softmax(logits_nl / temperature, dim=-1)
        sorted_probs_nl, indices_nl = torch.sort(probs_nl, descending=True, dim=-1)
        cum_sum_probs_nl = torch.cumsum(sorted_probs_nl, dim=-1)
        top_p_nl = cum_sum_probs_nl < p
        top_p_nl[:, 1:] = top_p_nl[:, :-1].clone()
        top_p_nl[:, 0] = 1
        sorted_probs_nl[~top_p_nl] = 0.0
        sampled_sorted_indexes_n1 = torch.multinomial(sorted_probs_nl, 1)
        sample_b1 = torch.gather(indices_nl, -1, sampled_sorted_indexes_n1)
        sample_btc1 = sample_b1.view(*orig_shape, 1)
        if dim != -1:
            sample_btc1 = sample_btc1.transpose(dim, -1)
        return sample_btc1.squeeze(dim)


def top_k_sampling(logits_btc: Tensor, k: int, temperature: float = 1.0, dim: int = -1) -> Tensor:
    """Samples from a distribution using top-k sampling.

    This function is a modified version of ``torch.multinomial`` that uses
    top-k sampling instead of top-p sampling. The difference is that top-k
    sampling sets the probability of all values outside the top-k to zero,
    whereas top-p sampling sets the probability of all values outside the
    top-p to zero.

    Parameters:
        logits_btc: The input tensor, of shape ``(B, T, C)``.
        k: The number of top values to consider.
        temperature: The temperature to apply to the logits.
        dim: The dimension to sample from. Defaults to ``-1``.

    Returns:
        The sampled indices, of shape ``(B, T)``.
    """
    with torch.no_grad():
        if dim != -1:
            logits_btc = logits_btc.transpose(dim, -1)
        orig_shape = logits_btc.shape[:-1]
        logits_nl = logits_btc.flatten(0, -2)
        probs_nl = F.softmax(logits_nl / temperature, dim=-1)
        sorted_probs_nl, indices_nl = torch.sort(probs_nl, descending=True, dim=-1)
        sorted_probs_nl[:, k:] = 0.0
        sorted_probs_nl = sorted_probs_nl / sorted_probs_nl.sum(dim=-1, keepdim=True)
        sampled_sorted_indexes_n1 = torch.multinomial(sorted_probs_nl, 1)
        sample_b1 = torch.gather(indices_nl, -1, sampled_sorted_indexes_n1)
        sample_btc1 = sample_b1.view(*orig_shape, 1)
        if dim != -1:
            sample_btc1 = sample_btc1.transpose(dim, -1)
        return sample_btc1.squeeze(dim)


@overload
def sample_from_logits(
    logits_btc: Tensor,
    strategy: Literal["top-p"],
    *,
    p: float,
    temperature: float = 1.0,
) -> Tensor: ...


@overload
def sample_from_logits(
    logits_btc: Tensor,
    strategy: Literal["top-k"],
    *,
    k: int,
    temperature: float = 1.0,
) -> Tensor: ...


@overload
def sample_from_logits(logits_btc: Tensor, strategy: Literal["greedy"]) -> Tensor: ...


@overload
def sample_from_logits(
    logits_btc: Tensor,
    strategy: SamplingStrategy,
    *,
    k: int | None = None,
    p: float | None = None,
    temperature: float = 1.0,
) -> Tensor: ...


def sample_from_logits(
    logits_btc: Tensor,
    strategy: SamplingStrategy,
    *,
    k: int | None = None,
    p: float | None = None,
    temperature: float = 1.0,
) -> Tensor:
    """Samples from a distribution using a given strategy.

    This function is a wrapper around the various sampling strategies, such as
    top-p sampling, top-k sampling, and greedy sampling.

    Parameters:
        logits_btc: The input tensor, of shape ``(B, T, C)``.
        strategy: The sampling strategy to use.
        k: The number of top values to consider, for top-k sampling.
        p: The probability threshold, for top-p sampling.
        temperature: The temperature to apply to the logits.

    Returns:
        The sampled indices, of shape ``(B, T)``.
    """
    match strategy:
        case "top-p":
            if p is None:
                raise ValueError("Top-P sampling requires a probability threshold.")
            return top_p_sampling(logits_btc, p, temperature)
        case "top-k":
            if k is None:
                raise ValueError("Top-K sampling requires a number of top values to consider.")
            return top_k_sampling(logits_btc, k, temperature)
        case "greedy":
            return torch.argmax(logits_btc, dim=-1)
        case _:
            raise ValueError(f"Invalid sampling strategy: {strategy}")



"""Defines base configuration functions and utilities."""

import functools
import os
from dataclasses import dataclass, field as field_base
from pathlib import Path
from typing import Any, cast

import torch
from omegaconf import II, MISSING, Container as OmegaConfContainer, OmegaConf

from mlfab.utils.text import show_error

FieldType = Any


def field(value: FieldType, **kwargs: str) -> FieldType:
    """Short-hand function for getting a config field.

    Args:
        value: The current field's default value.
        kwargs: Additional metadata fields to supply.

    Returns:
        The dataclass field.
    """
    metadata: dict[str, Any] = {}
    metadata.update(kwargs)

    if hasattr(value, "__call__"):
        return field_base(default_factory=value, metadata=metadata)
    if value.__class__.__hash__ is None:
        return field_base(default_factory=lambda: value, metadata=metadata)
    return field_base(default=value, metadata=metadata)


def is_missing(cfg: Any, key: str) -> bool:  # noqa: ANN401
    """Utility function for checking if a config key is missing.

    This is for cases when you are using a raw dataclass rather than an
    OmegaConf container but want to treat them the same way.

    Args:
        cfg: The config to check
        key: The key to check

    Returns:
        Whether or not the key is missing a value in the config
    """
    if isinstance(cfg, OmegaConfContainer):
        if OmegaConf.is_missing(cfg, key):
            return True
        if OmegaConf.is_interpolation(cfg, key):
            try:
                getattr(cfg, key)
                return False
            except Exception:
                return True
    if getattr(cfg, key) is MISSING:
        return True
    return False


@dataclass
class ErrorHandling:
    enabled: bool = field(True, help="Is error handling enabled?")
    maximum_exceptions: int = field(10, help="Maximum number of errors to encounter")
    backoff_after: int = field(5, help="Start to do a sleeping backoff after this many exceptions")
    sleep_backoff: float = field(0.1, help="Sleep backoff amount")
    sleep_backoff_power: float = field(2.0, help="How much to multiply backoff for each successive exception")
    log_full_exception: bool = field(False, help="Log the full exception message for each exception")
    flush_exception_summary_every: int = field(500, help="How often to flush exception summary")
    report_top_n_exception_types: int = field(5, help="Number of exceptions to summarize")
    exception_location_traceback_depth: int = field(3, help="Traceback length for the exception location")


@dataclass
class Logging:
    hide_third_party_logs: bool = field(True, help="If set, hide third-party logs")
    log_level: str = field("INFO", help="The logging level to use")


@dataclass
class Device:
    cpu: bool = field(True, help="Whether to use the CPU")
    gpu: bool = field(II("oc.env:USE_GPU,1"), help="Whether to use the GPU")
    metal: bool = field(II("oc.env:USE_METAL,1"), help="Whether to use the Apple Silicon accelerator")
    use_fp64: bool = field(False, help="Always use the 64-bit floating point type")
    use_fp32: bool = field(False, help="Always use the 32-bit floating point type")
    use_bf16: bool = field(False, help="Always use the 16-bit bfloat type")
    use_fp16: bool = field(False, help="Always use the 16-bit floating point type")


def parse_dtype(cfg: Device) -> torch.dtype | None:
    if cfg.use_fp64:
        return torch.float64
    if cfg.use_fp32:
        return torch.float32
    if cfg.use_bf16:
        return torch.bfloat16
    if cfg.use_fp16:
        return torch.float16
    return None


@dataclass
class Triton:
    use_triton_if_available: bool = field(True, help="Use Triton if available")


@dataclass
class Experiment:
    default_random_seed: int = field(1337, help="The default random seed to use")
    max_workers: int = field(32, help="Maximum number of workers to use")
    multiprocessing_start_method: str | None = field("forkserver", help="The multiprocessing start method to use")


@dataclass
class Directories:
    run: str = field(II("oc.env:RUN_DIR"), help="The run directory")
    data: str = field(II("oc.env:DATA_DIR"), help="The data directory")
    pretrained_models: str = field(II("oc.env:MODEL_DIR"), help="The models directory")


@dataclass
class SlurmPartition:
    partition: str = field(MISSING, help="The partition name")
    num_nodes: int = field(1, help="The number of nodes to use")


@dataclass
class Slurm:
    launch: dict[str, SlurmPartition] = field({}, help="The available launch configurations")


@dataclass
class UserConfig:
    error_handling: ErrorHandling = field(ErrorHandling)
    logging: Logging = field(Logging)
    device: Device = field(Device)
    triton: Triton = field(Triton)
    experiment: Experiment = field(Experiment)
    directories: Directories = field(Directories)
    slurm: Slurm = field(Slurm)


def user_config_path() -> Path:
    mlfabrc_path_raw = os.environ.get("MLFABRC_PATH", "~/.mlfab.yml")
    mlfabrc_path = Path(mlfabrc_path_raw).expanduser()
    return mlfabrc_path


@functools.lru_cache(maxsize=None)
def _load_user_config_cached() -> UserConfig:
    mlfabrc_path = user_config_path()
    base_cfg = OmegaConf.structured(UserConfig)

    # Writes the config file.
    if mlfabrc_path.exists():
        cfg = OmegaConf.merge(base_cfg, OmegaConf.load(mlfabrc_path))
    else:
        show_error(f"No config file was found in {mlfabrc_path}; writing one...", important=True)
        OmegaConf.save(base_cfg, mlfabrc_path)
        cfg = base_cfg

    # Looks in the current directory for a config file.
    local_cfg_path = Path("mlfab.yml")
    if local_cfg_path.exists():
        cfg = OmegaConf.merge(cfg, OmegaConf.load(local_cfg_path))

    return cast(UserConfig, cfg)


def load_user_config() -> UserConfig:
    """Loads the ``~/.mlfab.yml`` configuration file.

    Returns:
        The loaded configuration.
    """
    return _load_user_config_cached()


def get_run_dir() -> Path | None:
    config = load_user_config().directories
    if is_missing(config, "run"):
        return None
    (run_dir := Path(config.run)).mkdir(parents=True, exist_ok=True)
    return run_dir


def get_data_dir() -> Path:
    config = load_user_config().directories
    if is_missing(config, "data"):
        raise RuntimeError(
            "The data directory has not been set! You should set it in your config file "
            f"in {user_config_path()} or set the DATA_DIR environment variable."
        )
    return Path(config.data)


def get_pretrained_models_dir() -> Path:
    config = load_user_config().directories
    if is_missing(config, "pretrained_models"):
        raise RuntimeError(
            "The data directory has not been set! You should set it in your config file "
            f"in {user_config_path()} or set the MODEL_DIR environment variable."
        )
    return Path(config.pretrained_models)



"""Defines a logger which logs the current training state."""

from pathlib import Path
from typing import Literal

from omegaconf import DictConfig, OmegaConf

from mlfab.task.logger import LoggerImpl, LogLine


class StateLogger(LoggerImpl):
    def __init__(
        self,
        run_directory: str | Path,
        git_state_name: str = "git_state.txt",
        train_code_name: str = "train_code.py",
        config_name: str = "config.yaml",
        flush_immediately: bool = False,
        open_mode: Literal["w", "a"] = "w",
        line_sep: str = "\n",
        remove_unicode_from_namespaces: bool = True,
    ) -> None:
        super().__init__(float("inf"))

        self.git_state_file = Path(run_directory).expanduser().resolve() / git_state_name
        self.train_code_file = Path(run_directory).expanduser().resolve() / train_code_name
        self.config_file = Path(run_directory).expanduser().resolve() / config_name
        self.flush_immediately = flush_immediately
        self.open_mode = open_mode
        self.line_sep = line_sep
        self.remove_unicode_from_namespaces = remove_unicode_from_namespaces

    def log_git_state(self, git_state: str) -> None:
        with open(self.git_state_file, "w") as f:
            f.write(git_state)

    def log_training_code(self, training_code: str) -> None:
        with open(self.train_code_file, "w") as f:
            f.write(training_code)

    def log_config(self, config: DictConfig) -> None:
        OmegaConf.save(config, self.config_file)

    def write(self, line: LogLine) -> None:
        pass



"""Defines the base learning rate scheduler.

Here's an example of building a custom learning rate scheduler using this API
and using it in your own workflow:

.. code-block:: python

    class ConstantLRScheduler(BaseLRScheduler):
        def get_lr_sclae(self, state: State) -> float:
            return 1.0

    opt = Adam(model.parameters(), lr=1e-3)
    sched = ConstantLRScheduler.get(opt)

    state = State.init_state()

    for sample in samples:
        opt.zero_grad()
        loss = model(sample)
        loss.backward()
        opt.step()
        sched.step(state)
        state.num_steps += 1

Note that each scheduler expects to recieve a State dataclass as input.
"""

import math
from abc import ABC, abstractmethod
from typing import Any, Sequence

from torch import nn
from torch.optim.optimizer import Optimizer

from mlfab.core.state import State


class SchedulerAdapter:
    """Defines a general-purpose learning rate scheduler adapter."""

    last_state: State | None

    def __init__(self, scheduler: "BaseLRScheduler", optimizer: Optimizer) -> None:
        self.scheduler = scheduler
        self.optimizer = optimizer
        self.last_state = None

        for param_group in self.optimizer.param_groups:
            param_group["initial_lr"] = param_group["lr"]

        self.lr_scale = 0.0

    def state_dict(self) -> dict[str, Any]:
        return self.optimizer.state_dict()

    def load_state_dict(self, state_dict: dict) -> None:
        self.optimizer.load_state_dict(state_dict)

    def step(self, state: State) -> None:
        self.last_state = state
        self.lr_scale = self.scheduler.get_lr_scale(state)
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = param_group["initial_lr"] * self.lr_scale

    def parameters(self) -> Sequence[nn.Parameter]:
        return [p for param_group in self.optimizer.param_groups for p in param_group["params"]]


class BaseLRScheduler(ABC):
    """Defines the base learning rate scheduler."""

    @abstractmethod
    def get_lr_scale(self, state: State) -> float:
        """Given a state, returns the current learning rate.

        Args:
            state: The current trainer state

        Returns:
            The computed learning rate to use
        """

    def get(self, optimizer: Optimizer) -> SchedulerAdapter:
        return SchedulerAdapter(self, optimizer)


class ConstantLRScheduler(BaseLRScheduler):
    """Defines a constant learning rate scheduler.

    This scheduler is really only meant to be used as a quick stand-in for
    other schedulers when none is desired, or as an alternative method of
    sweeping the learning rate besides updating the optimizer.

    Parameters:
        factor: The constant learning rate factor.
    """

    def __init__(self, factor: float = 1.0) -> None:
        super().__init__()

        self.factor = factor

    def get_lr_scale(self, state: State) -> float:
        return self.factor


class LinearLRScheduler(BaseLRScheduler):
    """Ramps up to some value, then ramps back down.

    Parameters:
        total_steps: The total number of training steps.
        warmup_steps: The number of warmup steps.
        warmup_percent: The percent of total steps to use as warmup steps,
            if not specified.
        min_scale: The minimum learning rate scale.
        decay: Whether to decay the learning rate after warmup.
    """

    def __init__(
        self,
        total_steps: int | None = None,
        warmup_steps: int | None = None,
        warmup_percent: float = 0.01,
        min_scale: float = 1e-4,
        decay: bool = True,
    ) -> None:
        super().__init__()

        if warmup_steps is None:
            if total_steps is None:
                raise ValueError(
                    "If `total_steps` is not specified, then `warmup_steps` cannot be inferred from `warmup_percent`. "
                    "You should therefore specify the number of warmup steps explicitly."
                )
            warmup_steps = round(total_steps * warmup_percent)

        self.total_steps = total_steps
        self.warmup_steps = warmup_steps
        self.min_scale = min_scale
        self.decay = decay

    def get_lr_scale(self, state: State) -> float:
        warmup, total, min_scale = self.warmup_steps, self.total_steps, self.min_scale
        if state.num_steps < warmup:
            return state.num_steps / warmup
        if not self.decay:
            return 1.0
        if total is None:
            return 1.0
        if state.num_steps < total:
            return (1 - min_scale) * (total - state.num_steps) / (total - warmup) + min_scale
        return min_scale


class CosineLRScheduler(BaseLRScheduler):
    """Defines a cosine learning rate schedule.

    Parameters:
        total_steps: The total number of training steps.
        num_resets: The number of times to reset the learning rate.
        phase: The number of steps in a phase.
        ramp_up_percent: The percent of phase to spend ramping up.
        ramp_up_steps: The number of steps to spend ramping up.
        eta_min: The minimum learning rate scale.
        eta_max: The maximum learning rate scale.
    """

    def __init__(
        self,
        total_steps: int | None = None,
        num_resets: int = 0,
        phase: int | None = None,
        ramp_up_percent: float = 0.05,
        ramp_up_steps: int | None = None,
        eta_min: float = 0.01,
        eta_max: float = 1.0,
    ) -> None:
        super().__init__()

        if phase is None:
            if total_steps is None:
                raise ValueError("If `total_steps` is not specified, then `phase` must be specified.")
            phase = int(total_steps / (num_resets + 1))
        if ramp_up_steps is None:
            assert 0.0 <= ramp_up_percent < 1.0
            ramp_up_steps = round(phase * ramp_up_percent)
        else:
            assert ramp_up_steps < phase

        self.phase = phase
        self.ramp_up_steps = ramp_up_steps
        self.eta_min = eta_min
        self.eta_max = eta_max

    def get_lr_scale(self, state: State) -> float:
        phase, ramp_up = self.phase, self.ramp_up_steps
        eta_min, eta_max = self.eta_min, self.eta_max
        phase_steps = state.num_steps % (phase + ramp_up)
        if phase_steps < ramp_up:
            return (1.0 - eta_min) * (phase_steps / ramp_up) + eta_min
        sigma = (phase_steps - ramp_up) / phase
        return eta_min + (eta_max - eta_min) * (1 + math.cos(math.pi * sigma)) / 2


class CosineDecayLRScheduler(BaseLRScheduler):
    """Defines a cosine decay learning rate schedule.

    This is like a cosine learning rate scheduler, but it decays the learning
    rate over time. When using multiple resets, each reset will peak lower than
    the previous one.

    Parameters:
        total_steps: The total number of training steps.
        num_resets: The number of times to reset the learning rate.
        phase: The number of steps in a phase.
        ramp_up_percent: The percent of phase to spend ramping up.
        ramp_up_steps: The number of steps to spend ramping up.
        eta_min: The minimum learning rate scale.
        eta_max: The maximum learning rate scale.
        min_decay: The minimum learning rate decay.
    """

    def __init__(
        self,
        total_steps: int,
        num_resets: int = 0,
        phase: int | None = None,
        ramp_up_percent: float = 0.05,
        ramp_up_steps: int | None = None,
        eta_min: float = 0.01,
        eta_max: float = 1.0,
        min_decay: float = 1e-4,
    ) -> None:
        super().__init__()

        if phase is None:
            phase = int(total_steps / (num_resets + 1))
        if ramp_up_steps is None:
            assert 0.0 <= ramp_up_percent < 1.0
            ramp_up_steps = round(phase * ramp_up_percent)
        else:
            assert ramp_up_steps < phase

        self.phase = phase
        self.total_steps = total_steps
        self.ramp_up_steps = ramp_up_steps
        self.eta_min = eta_min
        self.eta_max = eta_max
        self.min_decay = min_decay

    def get_lr_scale(self, state: State) -> float:
        phase, total, ramp_up = self.phase, self.total_steps, self.ramp_up_steps
        eta_min, eta_max, min_decay = self.eta_min, self.eta_max, self.min_decay
        decay = (1.0 - min_decay) * (total - state.num_steps) / total + min_decay
        phase_steps = state.num_steps % (phase + ramp_up)
        if phase_steps < ramp_up:
            return max(phase_steps / ramp_up, eta_min) * decay
        sigma = (phase_steps - ramp_up) / phase
        lr_scale_no_decay = eta_min + (eta_max - eta_min) * (1 + math.cos(math.pi * sigma)) / 2
        return lr_scale_no_decay * decay


class NoamLRScheduler(BaseLRScheduler):
    """Defines the Noam learning rate scheduler.

    This corresponds to increasing the learning rate linearly for the first
    ``warmup_steps`` training steps, and decreasing it thereafter proportionally
    to the inverse square root of the step number, scaled by the inverse square
    root of the dimensionality of the model.

    Parameters:
        warmup_steps: The number of learning rate ramp up steps.
        model_size: The dimensionality of the model.
    """

    def __init__(self, warmup_steps: int, model_size: int = 1024) -> None:
        super().__init__()

        self.warmup_steps = warmup_steps
        self.model_size = model_size

    def get_lr_scale(self, state: State) -> float:
        return self.model_size**-0.5 * min(state.num_steps**-0.5, state.num_steps * self.warmup_steps**-1.5)



"""A task mixin to support ``torch.compile``."""

import logging
from dataclasses import dataclass
from typing import Callable, Generic, ParamSpec, TypeVar, cast

import torch
from omegaconf import II
from torch import nn

from mlfab.core.conf import field
from mlfab.task.mixins.device import DeviceConfig, DeviceMixin

logger = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")

Model = TypeVar("Model", bound=nn.Module)


@dataclass
class TorchCompileOptions:
    model: bool = field(II("oc.env:COMPILE_MODEL,0"), help="Enable Torch compilation for the model")
    func: bool = field(II("oc.env:COMPILE_FUNC,0"), help="Enable Torch compilation for functions")
    fullgraph: bool = field(False, help="Whether it is OK to break the model into subgraphs")
    dynamic: bool = field(False, help="Whether to use dynamic shape tracing")
    backend: str = field("auto", help="The backend to use")
    model_mode: str | None = field("max-autotune", help="Either 'default', 'reduce-overhead' or 'max-autotune'")
    func_mode: str | None = field("reduce-overhead", help="Either 'default', 'reduce-overhead' or 'max-autotune'")


@dataclass
class CompileConfig(DeviceConfig):
    compiler: TorchCompileOptions = field(TorchCompileOptions(), help="Torch compile config")


Config = TypeVar("Config", bound=CompileConfig)


class CompileMixin(DeviceMixin[Config], Generic[Config]):
    """Defines a mixin for calling `torch.compile` on models and functions."""

    def get_compiler_backend(self) -> str | Callable:
        backend: str | Callable = self.config.compiler.backend
        if backend == "auto":
            backend = self.device.get_torch_compile_backend()
            logger.info("Using torch-compile backend [%s]", backend)
        return backend

    def compile_model(self, model: Model) -> Model:
        if self.config.compiler.model:
            model = cast(
                Model,
                torch.compile(
                    model,
                    fullgraph=self.config.compiler.fullgraph,
                    dynamic=self.config.compiler.dynamic,
                    backend=self.get_compiler_backend(),
                    mode=self.config.compiler.model_mode,
                    disable=not self.config.compiler.model,
                ),
            )

        return model

    def compile_function(self, func: Callable[P, T]) -> Callable[P, T]:
        if self.config.compiler.func:
            func = torch.compile(
                func,
                fullgraph=self.config.compiler.fullgraph,
                dynamic=self.config.compiler.dynamic,
                backend=self.get_compiler_backend(),
                mode=self.config.compiler.func_mode,
                disable=not self.config.compiler.func,
            )

        return func



"""Defines a general-purpose API for weight initialization.

.. highlight:: python
.. code-block:: python

    from mlfab.nn.init import init_, cast_init_type

    linear = nn.Linear(32, 32)
    init_(linear.weight, linear.bias, "orthogonal")

    # This lets you parametrize the initialization type as a string.
    init_(linear.weight, linear.bias, cast_init_type(my_init_type))

Choices for the initialization type are:

- ``"orthogonal"``: Orthogonal initialization, meaning that the weights are initialized to an orthogonal matrix.
- ``"normal"``: Initializes weights with a normal distribution
- ``"biased_normal"``: Initializes both weights and biases with a normal distribution
- ``"uniform"``: Initializes weights with a uniform distribution
- ``"kaiming_uniform"`` or ``"kaiming_normal"``: Initializes weights with a Kaiming normal or uniform distribution
- ``"xavier_uniform"`` or ``"xavier_normal"``: Initializes weights with a Xavier normal or uniform distribution
- ``"zeros"``: Initializes weights to all zeros
- ``"ones"``: Initializes weights to all ones
"""

import math
from typing import Literal, cast, get_args

import torch
from torch import Tensor, nn

InitializationType = Literal[
    "orthogonal",
    "normal",
    "biased_normal",
    "uniform",
    "kaiming_uniform",
    "kaiming_normal",
    "xavier_uniform",
    "xavier_normal",
    "trunc_normal",
    "dirac",
    "constant",
    "zeros",
    "ones",
]


def cast_init_type(s: str) -> InitializationType:
    args = get_args(InitializationType)
    assert s in args, f"Invalid initialization type: '{s}' Valid options are {args}"
    return cast(InitializationType, s)


def _uniform_bias(weight: Tensor, bias: Tensor | None) -> Tensor | None:
    if bias is None:
        return None
    fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weight)
    if fan_in == 0:
        nn.init.zeros_(bias)
    else:
        bound = 1 / math.sqrt(fan_in)
        nn.init.uniform_(bias, -bound, bound)
    return bias


def _zeros(t: Tensor | None) -> Tensor | None:
    return None if t is None else nn.init.zeros_(t)


def init_(
    weight: Tensor,
    bias: Tensor | None,
    init: InitializationType,
    *,
    mean: float = 0.0,
    std: float = 0.01,
    scale: float = 0.02,
    groups: int = 1,
    trunc_clip: tuple[float, float] = (-2.0, 2.0),
) -> tuple[Tensor, Tensor | None]:
    """Initializes the weight and bias in-place, using an initialization key.

    The weight and bias are from a convolution or linear layer.

    Args:
        weight: The weight tensor
        bias: The bias tensor
        init: The initialization type to use
        mean: The mean for normal initialization
        std: The standard deviation for normal initialization
        scale: The scale amount for uniform or constant initialization
        groups: The number of groups, if argument is necessary
        trunc_clip: The min and max values for trunc_normal initialization

    Returns:
        The initialized weight and bias (which can be discarded, since the
        initialization happens in-place).

    Raises:
        NotImplementedError: If the initialization mode isn't implemented
    """
    # Don't do anything for meta tensors.
    if weight.is_meta:
        return weight, bias
    if isinstance(weight, nn.Parameter):
        weight = weight.data
    if isinstance(bias, nn.Parameter):
        bias = bias.data
    match init:
        case "orthogonal":
            if weight.dtype in (torch.float16, torch.bfloat16):
                return (
                    weight.copy_(nn.init.orthogonal_(weight.float(), gain=0.01).to(weight)),
                    _zeros(bias),
                )
            return nn.init.orthogonal_(weight), _zeros(bias)
        case "normal":
            return nn.init.normal_(weight, mean=mean, std=std), _zeros(bias)
        case "biased_normal":
            return (
                nn.init.normal_(weight, mean=mean, std=std),
                None if bias is None else nn.init.normal_(bias, mean=mean, std=std),
            )
        case "uniform":
            return nn.init.uniform_(weight, b=scale), _zeros(bias)
        case "kaiming_uniform":
            return nn.init.kaiming_uniform_(weight), _uniform_bias(weight, bias)
        case "kaiming_normal":
            return nn.init.kaiming_normal_(weight), _uniform_bias(weight, bias)
        case "xavier_uniform":
            return nn.init.xavier_uniform_(weight), _uniform_bias(weight, bias)
        case "xavier_normal":
            return nn.init.xavier_normal_(weight), _uniform_bias(weight, bias)
        case "trunc_normal":
            a, b = trunc_clip
            return nn.init.trunc_normal_(weight, mean=mean, std=std, a=a, b=b), _zeros(bias)
        case "dirac":
            return nn.init.dirac_(weight, groups=groups), _zeros(bias)
        case "constant":
            return nn.init.constant_(weight, scale), _zeros(bias)
        case "zeros":
            return nn.init.zeros_(weight), _zeros(bias)
        case "ones":
            return nn.init.ones_(weight), _zeros(bias)
        case _:
            raise NotImplementedError(f"Unexpected initialization: {init}")



"""Defines a mixin which provides a "run" method."""

import signal
from abc import ABC, abstractmethod
from dataclasses import dataclass
from types import FrameType
from typing import Callable, TypeVar

from mlfab.task.base import BaseConfig, BaseTask, RawConfigType
from mlfab.task.launchers.base import BaseLauncher


@dataclass
class RunnableConfig(BaseConfig):
    pass


Config = TypeVar("Config", bound=RunnableConfig)


class RunnableMixin(BaseTask[Config], ABC):
    def __init__(self, config: Config) -> None:
        super().__init__(config)

        self.__signal_handlers: dict[signal.Signals, list[Callable[[], None]]] = {}
        self.__set_signal_handlers: set[signal.Signals] = set()

    @abstractmethod
    def run(self) -> None:
        """Runs the task."""

    @classmethod
    def launch(
        cls,
        *cfgs: RawConfigType,
        launcher: BaseLauncher | None = None,
        use_cli: bool | list[str] = True,
    ) -> None:
        if launcher is None:
            from mlfab.task.launchers.cli import CliLauncher

            launcher = CliLauncher()
        launcher.launch(cls, *cfgs, use_cli=use_cli)

    def call_signal_handler(self, sig: int | signal.Signals, frame: FrameType | None = None) -> None:
        if isinstance(sig, int):
            sig = signal.Signals(sig)
        for signal_handler in self.__signal_handlers.get(sig, []):
            signal_handler()

    def add_signal_handler(self, handler: Callable[[], None], *sigs: signal.Signals) -> None:
        for sig in sigs:
            if sig not in self.__signal_handlers:
                self.__signal_handlers[sig] = []
            if sig not in self.__set_signal_handlers:
                self.__set_signal_handlers.add(sig)
                signal.signal(sig, self.call_signal_handler)
            self.__signal_handlers[sig].append(handler)



"""Defines a general-purpose API for transformer embedding layers.

.. highlight:: python
.. code-block:: python

    from mlfab.nn.embeddings import get_positional_embeddings, cast_embedding_kind

    embeddings = get_positional_embeddings(
        max_tsz=1024,
        embed_dim=128,
        kind="sinusoidal",
        learnable=False,
    )

    x = torch.arange(3, 5, 8)

    # Time-based positional embeddings - the time tensor supplies the
    # times for each element in the input.
    times = torch.randint(0, 1024, (3, 5))
    y1 = embeddings(x, times=times)

    # Offset-based positional embeddings - the input is assumed to be in
    # temporal order, and the offset is the offset of the first element.
    y2 = embeddings(x, offset=1)

    assert y1.shape == y2.shape == x.shape

    # This lets you parametrize the embedding kind as a string.
    embeddings = get_positional_embeddings(..., kind=cast_embedding_kind(my_kind))

Choices for the embedding kind are:

- ``"identity"``: No positional embeddings are added.
- ``"learned"``: Positional embeddings are learned.
- ``"sinusoidal"``: Sinusoidal embeddings.
- ``"rotary"``: Rotary embeddings (popular for training transformers).
"""

import math
from typing import Literal, cast, get_args, overload

import torch
from torch import Tensor, nn

from mlfab.nn.init import InitializationType, init_

EmbeddingKind = Literal["identity", "learned", "sinusoidal", "rotary"]


def cast_embedding_kind(k: str) -> EmbeddingKind:
    args = get_args(EmbeddingKind)
    assert k in args, f"Invalid initialization type: '{k}' Valid options are {args}"
    return cast(EmbeddingKind, k)


class IdentityPositionalEmbeddings(nn.Module):
    def forward(self, x: Tensor, offset: int = 0, times_bt: Tensor | None = None) -> Tensor:
        return x


class LearnedPositionalEmbeddings(nn.Module):
    """Defines a learned embeddings module.

    Parameters:
        max_tsz: The maximum sequence length.
        embed_dim: The embedding dimension.
        weight_init: The initialization type for the embedding weight.
        learnable: Whether the embeddings are learnable.
    """

    def __init__(
        self,
        max_tsz: int,
        embed_dim: int,
        weight_init: InitializationType = "normal",
        learnable: bool = True,
    ) -> None:
        super().__init__()

        self.max_tsz = max_tsz
        self.embed_dim = embed_dim
        self.weight_init = weight_init

        self.embeddings_tc = nn.Parameter(torch.empty(max_tsz, embed_dim), requires_grad=learnable)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        init_(self.embeddings_tc.data, None, self.weight_init)

    def forward(self, x: Tensor, offset: int = 0, times_bt: Tensor | None = None) -> Tensor:
        emb_btc = (
            self.embeddings_tc[None, offset : offset + x.size(1)] if times_bt is None else self.embeddings_tc[times_bt]
        )
        return x + emb_btc


class SinusoidalEmbeddings(nn.Module):
    """Defines a sinusoidal embeddings module.

    Parameters:
        embed_dim: The embedding dimension.
        max_tsz: The maximum sequence length.
        learnable: Whether the embeddings are learnable.
        base: The base for the sinusoidal embeddings.
    """

    def __init__(
        self,
        embed_dim: int | None = None,
        max_tsz: int | None = None,
        learnable: bool = True,
        base: int = 10_000,
    ) -> None:
        super().__init__()

        self.max_tsz = max_tsz
        self.embed_dim = embed_dim
        self.base = base

        self.embeddings_tc: nn.Parameter | None = None
        if learnable:
            assert max_tsz is not None, "Learnable parameters require `max_tsz` to be set"
            assert embed_dim is not None, "Learnable parameters require `embed_dim` to be set"
            self.embeddings_tc = nn.Parameter(torch.empty(max_tsz, embed_dim), requires_grad=learnable)
            self.reset_parameters()

        self.embeddings_cached: Tensor | None = None

    def forward(self, x_btc: Tensor, offset: int = 0, times_bt: Tensor | None = None) -> Tensor:
        embeddings_tc: Tensor | None = self.embeddings_tc
        _, tsz, xdim = x_btc.shape
        if embeddings_tc is None:
            max_tsz = max(tsz, 0 if times_bt is None else int(times_bt.max().item()) + 1) + offset
            if self.embeddings_cached is None:
                self.embeddings_cached = self.get_embeddings(max_tsz, xdim, x_btc.device, x_btc.dtype)
            else:
                embed_tsz, embed_dim = self.embeddings_cached.shape
                embed_device, embed_dtype = self.embeddings_cached.device, self.embeddings_cached.dtype
                if (
                    embed_tsz < max_tsz
                    or embed_dim != xdim
                    or embed_device != x_btc.device
                    or embed_dtype != x_btc.dtype
                ):
                    self.embeddings_cached = self.get_embeddings(max_tsz, embed_dim, x_btc.device, x_btc.dtype)
            embeddings_tc = self.embeddings_cached
        return x_btc + (embeddings_tc[None, offset : offset + tsz] if times_bt is None else embeddings_tc[times_bt])

    def reset_parameters(self) -> None:
        if self.embeddings_tc is None:
            assert self.max_tsz is not None, "Learnable parameters require `max_tsz` to be set"
            assert self.embed_dim is not None, "Learnable parameters require `embed_dim` to be set"
            self.embeddings_tc.data.copy_(self.get_embeddings(self.max_tsz, self.embed_dim))

    def get_embeddings(
        self,
        tsz: int,
        embed_dim: int,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> Tensor:
        positions = torch.arange(tsz, device=device, dtype=torch.float32)
        dim = torch.arange(embed_dim, device=device, dtype=torch.float32)
        dim = self.base ** (2 * (dim // 2) / embed_dim)
        embeddings = positions[:, None] / dim[None, :]
        embeddings[:, 0::2] = torch.sin(embeddings[:, 0::2])
        embeddings[:, 1::2] = torch.cos(embeddings[:, 1::2])
        return embeddings.to(dtype)


def get_rotary_embeddings(
    tsz: int,
    embed_dim: int,
    device: torch.device,
    dtype: torch.dtype,
    offset: int = 0,
    base: int = 10_000,
) -> Tensor:
    with torch.no_grad():
        assert embed_dim % 4 == 0, f"Embedding dimension must be divisible by 4, got {embed_dim}"
        half_d = embed_dim // 2
        theta = 1.0 / (base ** (torch.arange(0, half_d, 2, device=device, dtype=torch.float32) / half_d))
        seq_idx = torch.arange(offset, tsz + offset, device=device, dtype=torch.float32)
        idx_theta_tc = torch.einsum("t,c->tc", seq_idx, theta)
        idx_theta2_tc = torch.cat([idx_theta_tc, idx_theta_tc], dim=1)
        cos_tc, sin_tc = idx_theta2_tc.cos(), idx_theta2_tc.sin()
        emb_2tc = torch.stack((cos_tc, sin_tc), dim=0).to(dtype)
        return emb_2tc


def apply_rotary_embeddings(x_btc: Tensor, embs_2tc: Tensor, offset: int = 0, times_bt: Tensor | None = None) -> Tensor:
    cos_tc, sin_tc = embs_2tc.unbind(0)
    _, tsz, embed_dim = x_btc.shape
    half_d = embed_dim // 2
    quarter_d = embed_dim // 4
    x_rope_btc, x_pass_btc = x_btc[..., :half_d], x_btc[..., half_d:]
    neg_half_x_btc = torch.cat([-x_rope_btc[..., quarter_d:], x_rope_btc[..., :quarter_d]], dim=-1)
    cos_part_btc = cos_tc[None, offset : offset + tsz] if times_bt is None else cos_tc[times_bt]
    sin_part_btc = sin_tc[None, offset : offset + tsz] if times_bt is None else sin_tc[times_bt]
    x_rope_btc = x_rope_btc * cos_part_btc + neg_half_x_btc * sin_part_btc
    return torch.cat((x_rope_btc, x_pass_btc), dim=-1)


def rotary_embeddings(x_btc: Tensor, offset: int = 0, base: int = 10_000) -> Tensor:
    """Defines a single function for applying rotary embeddings.

    This is slower than using the module, but it doesn't require
    pre-initializing the embeddings, so it can be used when running online.

    Args:
        x_btc: The input tensor, with shape ``(batch, tsz, embed_dim)``.
        offset: The offset for the first element.
        base: The base for the sinusoidal embeddings.

    Returns:
        The input tensor with rotary embeddings applied.
    """
    (_, tsz, embed_dim), device, dtype = x_btc.shape, x_btc.device, x_btc.dtype
    emb_2tc = get_rotary_embeddings(tsz + offset, embed_dim, device, dtype, 0, base)
    return apply_rotary_embeddings(x_btc, emb_2tc, offset)


class RotaryEmbeddings(nn.Module):
    def __init__(self, base: int = 10_000) -> None:
        """Defines a rotary embeddings module.

        Args:
            base: The base for the sinusoidal embeddings.
        """
        super().__init__()

        self.base = base
        self.embeddings: Tensor | None = None

    def forward(self, x_btc: Tensor, offset: int = 0, times_bt: Tensor | None = None) -> Tensor:
        emb_2tc = self.embeddings
        _, tsz, embed_dim = x_btc.shape
        max_tsz = max(tsz, 0 if times_bt is None else int(times_bt.max().item()) + 1) + offset
        if emb_2tc is None or emb_2tc.shape[-2] < max_tsz:
            emb_2tc = get_rotary_embeddings(max_tsz, embed_dim, x_btc.device, x_btc.dtype, 0, self.base)
            self.embeddings = emb_2tc
        return apply_rotary_embeddings(x_btc, emb_2tc, offset, times_bt)


@overload
def get_positional_embeddings(kind: Literal["identity"]) -> IdentityPositionalEmbeddings: ...


@overload
def get_positional_embeddings(
    kind: Literal["learned"],
    *,
    max_tsz: int,
    embed_dim: int,
    weight_init: InitializationType = "normal",
    learnable: bool | None = None,
) -> LearnedPositionalEmbeddings: ...


@overload
def get_positional_embeddings(
    kind: Literal["sinusoidal"],
    *,
    max_tsz: int | None = None,
    embed_dim: int | None = None,
    learnable: bool | None = None,
    base: int = 10_000,
) -> SinusoidalEmbeddings: ...


@overload
def get_positional_embeddings(
    kind: Literal["rotary"],
    *,
    base: int = 10_000,
) -> RotaryEmbeddings: ...


@overload
def get_positional_embeddings(
    kind: EmbeddingKind,
    *,
    max_tsz: int | None = None,
    embed_dim: int | None = None,
    weight_init: InitializationType = "normal",
    learnable: bool | None = None,
    base: int = 10_000,
) -> IdentityPositionalEmbeddings | LearnedPositionalEmbeddings | SinusoidalEmbeddings | RotaryEmbeddings: ...


def get_positional_embeddings(
    kind: EmbeddingKind,
    *,
    max_tsz: int | None = None,
    embed_dim: int | None = None,
    weight_init: InitializationType = "normal",
    learnable: bool | None = None,
    base: int = 10_000,
) -> nn.Module:
    """Defines the common module for adding positional embeddings.

    Args:
        kind: The type of embedding to use.
        max_tsz: The maximum sequence length.
        embed_dim: The embedding dimension.
        weight_init: The weight initialization for learned embeddings.
        learnable: Whether the embeddings are learnable; if not provided,
            uses sensible defaults.
        base: The base for the sinusoidal embeddings.

    Returns:
        The positional embeddings module.

    Raises:
        ValueError: If an invalid embedding kind is supplied.
    """
    match kind:
        case "identity":
            return IdentityPositionalEmbeddings()

        case "learned":
            assert max_tsz is not None, "Learned embeddings require `max_tsz` to be set"
            assert embed_dim is not None, "Learned embeddings require `embed_dim` to be set"

            return LearnedPositionalEmbeddings(
                max_tsz=max_tsz,
                embed_dim=embed_dim,
                weight_init=weight_init,
                learnable=True if learnable is None else learnable,
            )

        case "sinusoidal":
            return SinusoidalEmbeddings(
                max_tsz=max_tsz,
                embed_dim=embed_dim,
                learnable=False if learnable is None else learnable,
                base=base,
            )

        case "rotary":
            return RotaryEmbeddings(base=base)

        case _:
            raise ValueError(f"Invalid embedding kind: {kind}")


def fourier_embeddings(t: Tensor, dim: int, max_period: int = 10000) -> Tensor:
    half = dim // 2
    idxs = torch.arange(start=0, end=half, device=t.device, dtype=torch.float32)
    freqs = torch.exp(-math.log(max_period) * idxs / half)
    args = t[:, None].float() * freqs[None]
    embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
    # Adds an additional row of zeros to match the expected dimension.
    if dim % 2:
        embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    return embedding


class FourierEmbeddings(nn.Module):
    """Defines a module for applying Fourier embeddings to timesteps.

    This module differs from the other positional embedding modules because it
    expects a continuous time input, rather than a discrete time input.

    Parameters:
        dim: The number of embedding dimensions. This value is used to determine
            how many different frequencies to use, and a higher value means
            higher frequencies.
        max_period: The maximum period for the embeddings. This should roughly
            be in line with the maximum number of timesteps; the default value
            of 10,000 is commonly used in NLP applications, and is derived from
            operating on sequence lengths of 100 to 1000 tokens.
    """

    __constants__ = ["dim", "max_period"]

    def __init__(self, dim: int, max_period: int = 10000) -> None:
        super().__init__()

        self.dim = dim
        self.max_period = max_period

    def forward(self, t: Tensor) -> Tensor:
        return fourier_embeddings(t, self.dim, self.max_period)



"""Defines utilities for reading and writing data."""

import itertools
import math
from fractions import Fraction
from pathlib import Path
from typing import Iterator

import numpy as np
import torch
import torchvision.transforms.functional as V
from PIL import Image, ImageSequence
from torch import Tensor
from torchvision.transforms import InterpolationMode

VALID_CHANNEL_COUNTS = {1, 3}


def image_as_uint8(arr: np.ndarray) -> np.ndarray:
    if np.issubdtype(arr.dtype, np.integer):
        return arr.astype(np.uint8)
    if np.issubdtype(arr.dtype, np.floating):
        return (arr * 255).round().astype(np.uint8)
    raise NotImplementedError(f"Unsupported dtype: {arr.dtype}")


def make_human_viewable_resolution(
    image: Tensor,
    interpolation: InterpolationMode = InterpolationMode.BILINEAR,
    trg_res: tuple[int, int] = (250, 250),
) -> Tensor:
    """Resizes image to human-viewable resolution.

    Args:
        image: The image to resize, with shape (C, H, W)
        interpolation: Interpolation mode to use for image resizing
        trg_res: The target image resolution; the image will be reshaped to
            have approximately the same area as an image with this resolution

    Returns:
        The resized image
    """
    width, height = V.get_image_size(image)
    trg_height, trg_width = trg_res
    factor = math.sqrt((trg_height * trg_width) / (height * width))
    new_height, new_width = int(height * factor), int(width * factor)
    return V.resize(image, [new_height, new_width], interpolation)


def _aminmax(t: Tensor) -> tuple[Tensor, Tensor]:
    # `aminmax` isn't supported for MPS tensors, fall back to separate calls.
    minv, maxv = (t.min(), t.max()) if t.is_mps else tuple(t.aminmax())
    return minv, maxv


def standardize_image(
    image: np.ndarray | Tensor,
    *,
    log_key: str | None = None,
    normalize: bool = True,
    keep_resolution: bool = False,
) -> np.ndarray:
    """Converts an arbitrary image to shape (C, H, W).

    Args:
        image: The image tensor to log
        log_key: An optional logging key to use in the exception message
        normalize: Normalize images to (0, 1)
        keep_resolution: If set, preserve original image resolution, otherwise
            change image resolution to human-viewable

    Returns:
        The normalized image, with shape (H, W, C)

    Raises:
        ValueError: If the image shape is invalid
    """
    if isinstance(image, np.ndarray):
        image = torch.from_numpy(image)

    if normalize and image.is_floating_point():
        minv, maxv = _aminmax(image)
        maxv.clamp_min_(1.0)
        minv.clamp_max_(0.0)
        image = torch.clamp((image.detach() - minv) / (maxv - minv), 0.0, 1.0)

    if image.ndim == 2:
        image = image.unsqueeze(0)
    elif image.ndim == 3:
        if image.shape[0] in VALID_CHANNEL_COUNTS:
            pass
        elif image.shape[2] in VALID_CHANNEL_COUNTS:
            image = image.permute(2, 0, 1)
        else:
            raise ValueError(f"Invalid channel count{'' if log_key is None else f' for {log_key}'}: {image.shape}")
    else:
        raise ValueError(f"Invalid image shape{'' if log_key is None else f' for {log_key}'}: {image.shape}")

    if not keep_resolution:
        image = make_human_viewable_resolution(image)

    return image.permute(1, 2, 0).detach().cpu().numpy()


def read_gif(in_file: str | Path, *, skip_first_frame: bool = True) -> Iterator[np.ndarray]:
    """Function that reads a GIF and returns a stream of Numpy arrays.

    Args:
        in_file: The path to the input file.
        skip_first_frame: If set, skip the first frame.

    Yields:
        A stream of Numpy arrays with shape (H, W, C).
    """
    gif = Image.open(str(in_file))
    iterator = ImageSequence.Iterator(gif)
    if skip_first_frame:
        next(iterator)
    for frame in iterator:
        yield np.array(frame)


def write_gif(
    itr: Iterator[np.ndarray | Tensor],
    out_file: str | Path,
    *,
    keep_resolution: bool = False,
    fps: int | Fraction = 10,
    loop: bool = False,
    first_frame_zeros: bool = True,
) -> None:
    """Function that writes an GIF from a stream of input tensors.

    Args:
        itr: The image iterator, yielding images with shape (H, W, C).
        out_file: The path to the output file.
        keep_resolution: If set, preserve original image resolution, otherwise
            change image resolution to human-viewable.
        fps: Frames per second for the GIF.
        loop: If set, loop the GIF.
        first_frame_zeros: If set, the first frame will be all zeros.
    """

    def to_image(t: np.ndarray | Tensor) -> Image.Image:
        return Image.fromarray(standardize_image(t, keep_resolution=keep_resolution))

    first_frame = standardize_image(next(itr), keep_resolution=keep_resolution)
    first_img = Image.fromarray(np.zeros_like(first_frame) if first_frame_zeros else first_frame)
    first_img.save(
        str(out_file),
        save_all=True,
        append_images=itertools.chain((Image.fromarray(i) for i in (first_frame,)), (to_image(t) for t in itr)),
        duration=int(1000 / fps),  # Number of milliseconds per frame.
        loop=int(loop),
    )



"""Defines a dataclass for keeping track of the current training state."""

import time
from dataclasses import dataclass
from typing import Literal, TypeVar, cast, get_args

from omegaconf import MISSING
from torch import nn

from mlfab.core.conf import field

Module = TypeVar("Module", bound=nn.Module)

Phase = Literal["train", "valid", "test"]


def cast_phase(raw_phase: str) -> Phase:
    args = get_args(Phase)
    assert raw_phase in args, f"Invalid phase: '{raw_phase}' Valid options are {args}"
    return cast(Phase, raw_phase)


@dataclass
class State:
    num_steps: int = field(MISSING, help="Number of steps so far")
    num_samples: int = field(MISSING, help="Number of sample so far")
    num_valid_steps: int = field(MISSING, help="Number of validation steps so far")
    num_test_steps: int = field(MISSING, help="Number of test steps so far")
    start_time_s: float = field(MISSING, help="Start time of training")
    elapsed_time_s: float = field(MISSING, help="Total elapsed time so far")
    raw_phase: str = field(MISSING, help="Current training phase")

    @property
    def phase(self) -> Phase:
        return cast_phase(self.raw_phase)

    @phase.setter
    def phase(self, new_phase: Phase) -> None:
        self.raw_phase = new_phase

    @classmethod
    def init_state(cls) -> "State":
        return cls(
            num_steps=0,
            num_samples=0,
            num_valid_steps=0,
            num_test_steps=0,
            start_time_s=time.time(),
            elapsed_time_s=0.0,
            raw_phase="train",
        )

    @property
    def training(self) -> bool:
        return self.phase == "train"

    def num_phase_steps(self, phase: Phase) -> int:
        match phase:
            case "train":
                return self.num_steps
            case "valid":
                return self.num_valid_steps
            case "test":
                return self.num_test_steps

    def set_phase(self, model: Module, new_phase: Phase) -> Module:
        self.phase = new_phase
        if new_phase == "train":
            if not model.training:
                model = model.train()
            return model
        else:
            if model.training:
                model = model.eval()
            return model



"""Defines a utility function for detecting the training device."""

import logging

from mlfab.nn.device.base import base_device
from mlfab.nn.device.cpu import cpu_device
from mlfab.nn.device.gpu import gpu_device
from mlfab.nn.device.metal import metal_device
from mlfab.utils.logging import LOG_PING

logger: logging.Logger = logging.getLogger(__name__)

# Earlier devices in list will take precedence.
ALL_DEVICE_TYPES: list[type[base_device]] = [
    metal_device,
    gpu_device,
    cpu_device,
]


def detect_device() -> base_device:
    for device_type in ALL_DEVICE_TYPES:
        if device_type.has_device():
            device = device_type()
            logger.log(LOG_PING, f"Using device: {device}")
            return device
    raise RuntimeError("Could not automatically detect the device to use")



"""Defines the top-level mlfab API.

This package is structured so that all the important stuff can be accessed
without having to dig around through the internals. This is done by lazily
importing the module by name.

This file can be maintained by running the update script:

.. code-block:: bash

    python -m scripts.update_api --inplace
"""

__version__ = "0.1.9"

# This list shouldn't be modified by hand; instead, run the update script.
__all__ = [
    "UserConfig",
    "field",
    "get_data_dir",
    "get_pretrained_models_dir",
    "get_run_dir",
    "load_user_config",
    "State",
    "cast_phase",
    "Clamp",
    "Clamp6",
    "FastGELU",
    "LaplaceActivation",
    "QuickGELU",
    "ReLUSquared",
    "cast_activation_type",
    "get_activation",
    "MultiheadAttention",
    "NextTokenTransformer",
    "NextTokenWithEmbeddingsTransformer",
    "TransformerDecoder",
    "TransformerDecoderLayer",
    "TransformerEncoder",
    "TransformerEncoderLayer",
    "get_attention_mask",
    "BiFPN",
    "BiFPNBlock",
    "MonotonicAttention",
    "monotonic_attention",
    "sample_from_logits",
    "top_k_sampling",
    "top_p_sampling",
    "NextTokenRwkv",
    "NextTokenWithEmbeddingsRwkv",
    "RwkvAttention",
    "RwkvBlock",
    "RwkvFeedForward",
    "RwkvStack",
    "UNet",
    "detect_device",
    "base_device",
    "cpu_device",
    "gpu_device",
    "metal_device",
    "ConsistencyModel",
    "GaussianDiffusion",
    "cast_beta_schedule",
    "get_diffusion_beta_schedule",
    "BaseODESolver",
    "EulerODESolver",
    "HeunODESolver",
    "RK4ODESolver",
    "cast_solver_type",
    "get_ode_solver",
    "FourierEmbeddings",
    "IdentityPositionalEmbeddings",
    "LearnedPositionalEmbeddings",
    "RotaryEmbeddings",
    "SinusoidalEmbeddings",
    "apply_rotary_embeddings",
    "cast_embedding_kind",
    "fourier_embeddings",
    "get_positional_embeddings",
    "get_rotary_embeddings",
    "rotary_embeddings",
    "DropPath",
    "StreamingConv1d",
    "StreamingConvTranspose1d",
    "as_numpy_array",
    "as_tensor",
    "device_fn",
    "drop_path",
    "gated_residual",
    "recursive_apply",
    "recursive_chunk",
    "residual",
    "set_random_seed",
    "streamable_cbr",
    "streaming_add",
    "streaming_conv_1d",
    "streaming_conv_transpose_1d",
    "init_",
    "KMeans",
    "kmeans_fn",
    "LoraColumnParallelLinear",
    "LoraConv1d",
    "LoraConv2d",
    "LoraConvTranspose1d",
    "LoraConvTranspose2d",
    "LoraEmbedding",
    "LoraGRU",
    "LoraGRUCell",
    "LoraLSTM",
    "LoraLSTMCell",
    "LoraLinear",
    "LoraParallelEmbedding",
    "LoraRowParallelLinear",
    "freeze_non_lora_",
    "lora",
    "maybe_lora",
    "maybe_lora_weight_norm",
    "reset_lora_weights_",
    "ImageGradLoss",
    "LPIPS",
    "MFCCLoss",
    "MelLoss",
    "MultiResolutionSTFTLoss",
    "SSIMLoss",
    "STFTLoss",
    "get_stft_window",
    "kl_div_pair_loss",
    "kl_div_single_loss",
    "log_cosh_loss",
    "log_stft_magnitude_loss",
    "pseudo_huber_loss",
    "spectral_convergence_loss",
    "stft",
    "stft_magnitude_loss",
    "BaseLRScheduler",
    "ConstantLRScheduler",
    "CosineDecayLRScheduler",
    "CosineLRScheduler",
    "LinearLRScheduler",
    "SchedulerAdapter",
    "ConvLayerNorm",
    "LastBatchNorm",
    "cast_norm_type",
    "cast_parametrize_norm_type",
    "get_norm_1d",
    "get_norm_2d",
    "get_norm_3d",
    "get_norm_linear",
    "get_parametrization_norm",
    "Adam",
    "Lion",
    "can_use_foreach",
    "can_use_fused",
    "separate_decayable_params",
    "ColumnParallelLinear",
    "MultiProcessConfig",
    "ParallelConfig",
    "ParallelEmbedding",
    "RowParallelLinear",
    "ddp",
    "dp",
    "fsdp",
    "get_data_worker_info",
    "get_local_rank",
    "get_local_rank_optional",
    "get_local_world_size",
    "get_local_world_size_optional",
    "get_master_addr",
    "get_master_port",
    "get_rank",
    "get_rank_optional",
    "get_unused_port",
    "get_world_size",
    "get_world_size_optional",
    "is_distributed",
    "is_master",
    "launch_subprocesses",
    "split_n_items_across_workers",
    "FiniteScalarQuantization",
    "LookupFreeQuantization",
    "ResidualVectorQuantization",
    "VectorQuantization",
    "BaseLauncher",
    "CliLauncher",
    "MultiProcessLauncher",
    "SingleProcessLauncher",
    "SlurmLauncher",
    "LogAudio",
    "LogImage",
    "LogLine",
    "LogVideo",
    "Logger",
    "LoggerImpl",
    "make_square_image_or_video",
    "JsonLogger",
    "StateLogger",
    "StdoutLogger",
    "TensorboardLogger",
    "TorchCompileOptions",
    "CPUStatsOptions",
    "DataloaderConfig",
    "GPUStatsOptions",
    "ProfilerOptions",
    "Script",
    "ScriptConfig",
    "Config",
    "Task",
    "collate",
    "collate_nullable",
    "pad_all",
    "pad_sequence",
    "SmallDataset",
    "ErrorHandlingDataset",
    "ErrorHandlingIterableDataset",
    "ExceptionSummary",
    "error_handling_dataset",
    "SquareResizeCrop",
    "UpperLeftCrop",
    "denormalize",
    "make_same_size",
    "make_size",
    "pil_to_tensor",
    "random_square_crop",
    "random_square_crop_multi",
    "square_crop",
    "square_resize_crop",
    "upper_left_crop",
    "check_md5",
    "check_sha256",
    "ensure_downloaded",
    "get_git_state",
    "get_state_dict_prefix",
    "get_training_code",
    "save_config",
    "test_dataset",
    "read_gif",
    "write_gif",
    "ColoredFormatter",
    "configure_logging",
    "TextBlock",
    "colored",
    "format_datetime",
    "format_timedelta",
    "outlined",
    "render_text_blocks",
    "show_error",
    "show_warning",
    "uncolored",
    "wrapped",
    "TokenReader",
    "TokenWriter",
    "token_file",
]

__all__ += [
    "ActivationType",
    "CollateMode",
    "EmbeddingKind",
    "NormType",
    "ODESolverType",
    "OptType",
    "ParametrizationNormType",
    "Phase",
    "RwkvAttentionState",
    "RwkvFeedForwardState",
    "RwkvState",
]

import os
from typing import TYPE_CHECKING

# If this flag is set, eagerly imports the entire package (not recommended).
IMPORT_ALL = int(os.environ.get("MLFAB_IMPORT_ALL", "0")) != 0

del os

# This dictionary is auto-generated and shouldn't be modified by hand; instead,
# run the update script.
NAME_MAP: dict[str, str] = {
    "UserConfig": "core.conf",
    "field": "core.conf",
    "get_data_dir": "core.conf",
    "get_pretrained_models_dir": "core.conf",
    "get_run_dir": "core.conf",
    "load_user_config": "core.conf",
    "State": "core.state",
    "cast_phase": "core.state",
    "Clamp": "nn.activations",
    "Clamp6": "nn.activations",
    "FastGELU": "nn.activations",
    "LaplaceActivation": "nn.activations",
    "QuickGELU": "nn.activations",
    "ReLUSquared": "nn.activations",
    "cast_activation_type": "nn.activations",
    "get_activation": "nn.activations",
    "MultiheadAttention": "nn.architectures.attention",
    "NextTokenTransformer": "nn.architectures.attention",
    "NextTokenWithEmbeddingsTransformer": "nn.architectures.attention",
    "TransformerDecoder": "nn.architectures.attention",
    "TransformerDecoderLayer": "nn.architectures.attention",
    "TransformerEncoder": "nn.architectures.attention",
    "TransformerEncoderLayer": "nn.architectures.attention",
    "get_attention_mask": "nn.architectures.attention",
    "BiFPN": "nn.architectures.bifpn",
    "BiFPNBlock": "nn.architectures.bifpn",
    "MonotonicAttention": "nn.architectures.monotonic_attention",
    "monotonic_attention": "nn.architectures.monotonic_attention",
    "sample_from_logits": "nn.architectures.next_token",
    "top_k_sampling": "nn.architectures.next_token",
    "top_p_sampling": "nn.architectures.next_token",
    "NextTokenRwkv": "nn.architectures.rwkv",
    "NextTokenWithEmbeddingsRwkv": "nn.architectures.rwkv",
    "RwkvAttention": "nn.architectures.rwkv",
    "RwkvBlock": "nn.architectures.rwkv",
    "RwkvFeedForward": "nn.architectures.rwkv",
    "RwkvStack": "nn.architectures.rwkv",
    "UNet": "nn.architectures.unet",
    "detect_device": "nn.device.auto",
    "base_device": "nn.device.base",
    "cpu_device": "nn.device.cpu",
    "gpu_device": "nn.device.gpu",
    "metal_device": "nn.device.metal",
    "ConsistencyModel": "nn.diffusion.consistency",
    "GaussianDiffusion": "nn.diffusion.gaussian",
    "cast_beta_schedule": "nn.diffusion.gaussian",
    "get_diffusion_beta_schedule": "nn.diffusion.gaussian",
    "BaseODESolver": "nn.diffusion.ode",
    "EulerODESolver": "nn.diffusion.ode",
    "HeunODESolver": "nn.diffusion.ode",
    "RK4ODESolver": "nn.diffusion.ode",
    "cast_solver_type": "nn.diffusion.ode",
    "get_ode_solver": "nn.diffusion.ode",
    "FourierEmbeddings": "nn.embeddings",
    "IdentityPositionalEmbeddings": "nn.embeddings",
    "LearnedPositionalEmbeddings": "nn.embeddings",
    "RotaryEmbeddings": "nn.embeddings",
    "SinusoidalEmbeddings": "nn.embeddings",
    "apply_rotary_embeddings": "nn.embeddings",
    "cast_embedding_kind": "nn.embeddings",
    "fourier_embeddings": "nn.embeddings",
    "get_positional_embeddings": "nn.embeddings",
    "get_rotary_embeddings": "nn.embeddings",
    "rotary_embeddings": "nn.embeddings",
    "DropPath": "nn.functions",
    "StreamingConv1d": "nn.functions",
    "StreamingConvTranspose1d": "nn.functions",
    "as_numpy_array": "nn.functions",
    "as_tensor": "nn.functions",
    "device_fn": "nn.functions",
    "drop_path": "nn.functions",
    "gated_residual": "nn.functions",
    "recursive_apply": "nn.functions",
    "recursive_chunk": "nn.functions",
    "residual": "nn.functions",
    "set_random_seed": "nn.functions",
    "streamable_cbr": "nn.functions",
    "streaming_add": "nn.functions",
    "streaming_conv_1d": "nn.functions",
    "streaming_conv_transpose_1d": "nn.functions",
    "init_": "nn.init",
    "KMeans": "nn.kmeans",
    "kmeans_fn": "nn.kmeans",
    "LoraColumnParallelLinear": "nn.lora",
    "LoraConv1d": "nn.lora",
    "LoraConv2d": "nn.lora",
    "LoraConvTranspose1d": "nn.lora",
    "LoraConvTranspose2d": "nn.lora",
    "LoraEmbedding": "nn.lora",
    "LoraGRU": "nn.lora",
    "LoraGRUCell": "nn.lora",
    "LoraLSTM": "nn.lora",
    "LoraLSTMCell": "nn.lora",
    "LoraLinear": "nn.lora",
    "LoraParallelEmbedding": "nn.lora",
    "LoraRowParallelLinear": "nn.lora",
    "freeze_non_lora_": "nn.lora",
    "lora": "nn.lora",
    "maybe_lora": "nn.lora",
    "maybe_lora_weight_norm": "nn.lora",
    "reset_lora_weights_": "nn.lora",
    "ImageGradLoss": "nn.losses",
    "LPIPS": "nn.losses",
    "MFCCLoss": "nn.losses",
    "MelLoss": "nn.losses",
    "MultiResolutionSTFTLoss": "nn.losses",
    "SSIMLoss": "nn.losses",
    "STFTLoss": "nn.losses",
    "get_stft_window": "nn.losses",
    "kl_div_pair_loss": "nn.losses",
    "kl_div_single_loss": "nn.losses",
    "log_cosh_loss": "nn.losses",
    "log_stft_magnitude_loss": "nn.losses",
    "pseudo_huber_loss": "nn.losses",
    "spectral_convergence_loss": "nn.losses",
    "stft": "nn.losses",
    "stft_magnitude_loss": "nn.losses",
    "BaseLRScheduler": "nn.lr_schedulers",
    "ConstantLRScheduler": "nn.lr_schedulers",
    "CosineDecayLRScheduler": "nn.lr_schedulers",
    "CosineLRScheduler": "nn.lr_schedulers",
    "LinearLRScheduler": "nn.lr_schedulers",
    "SchedulerAdapter": "nn.lr_schedulers",
    "ConvLayerNorm": "nn.norms",
    "LastBatchNorm": "nn.norms",
    "cast_norm_type": "nn.norms",
    "cast_parametrize_norm_type": "nn.norms",
    "get_norm_1d": "nn.norms",
    "get_norm_2d": "nn.norms",
    "get_norm_3d": "nn.norms",
    "get_norm_linear": "nn.norms",
    "get_parametrization_norm": "nn.norms",
    "Adam": "nn.optimizers",
    "Lion": "nn.optimizers",
    "can_use_foreach": "nn.optimizers",
    "can_use_fused": "nn.optimizers",
    "separate_decayable_params": "nn.optimizers",
    "ColumnParallelLinear": "nn.parallel",
    "MultiProcessConfig": "nn.parallel",
    "ParallelConfig": "nn.parallel",
    "ParallelEmbedding": "nn.parallel",
    "RowParallelLinear": "nn.parallel",
    "ddp": "nn.parallel",
    "dp": "nn.parallel",
    "fsdp": "nn.parallel",
    "get_data_worker_info": "nn.parallel",
    "get_local_rank": "nn.parallel",
    "get_local_rank_optional": "nn.parallel",
    "get_local_world_size": "nn.parallel",
    "get_local_world_size_optional": "nn.parallel",
    "get_master_addr": "nn.parallel",
    "get_master_port": "nn.parallel",
    "get_rank": "nn.parallel",
    "get_rank_optional": "nn.parallel",
    "get_unused_port": "nn.parallel",
    "get_world_size": "nn.parallel",
    "get_world_size_optional": "nn.parallel",
    "is_distributed": "nn.parallel",
    "is_master": "nn.parallel",
    "launch_subprocesses": "nn.parallel",
    "split_n_items_across_workers": "nn.parallel",
    "FiniteScalarQuantization": "nn.quantization.fsq",
    "LookupFreeQuantization": "nn.quantization.lfq",
    "ResidualVectorQuantization": "nn.quantization.vq",
    "VectorQuantization": "nn.quantization.vq",
    "BaseLauncher": "task.launchers.base",
    "CliLauncher": "task.launchers.cli",
    "MultiProcessLauncher": "task.launchers.multi_process",
    "SingleProcessLauncher": "task.launchers.single_process",
    "SlurmLauncher": "task.launchers.slurm",
    "LogAudio": "task.logger",
    "LogImage": "task.logger",
    "LogLine": "task.logger",
    "LogVideo": "task.logger",
    "Logger": "task.logger",
    "LoggerImpl": "task.logger",
    "make_square_image_or_video": "task.logger",
    "JsonLogger": "task.loggers.json",
    "StateLogger": "task.loggers.state",
    "StdoutLogger": "task.loggers.stdout",
    "TensorboardLogger": "task.loggers.tensorboard",
    "TorchCompileOptions": "task.mixins.compile",
    "CPUStatsOptions": "task.mixins.cpu_stats",
    "DataloaderConfig": "task.mixins.data_loader",
    "GPUStatsOptions": "task.mixins.gpu_stats",
    "ProfilerOptions": "task.mixins.profiler",
    "Script": "task.script",
    "ScriptConfig": "task.script",
    "Config": "task.task",
    "Task": "task.task",
    "collate": "utils.data.collate",
    "collate_nullable": "utils.data.collate",
    "pad_all": "utils.data.collate",
    "pad_sequence": "utils.data.collate",
    "SmallDataset": "utils.data.dataset",
    "ErrorHandlingDataset": "utils.data.error_handling",
    "ErrorHandlingIterableDataset": "utils.data.error_handling",
    "ExceptionSummary": "utils.data.error_handling",
    "error_handling_dataset": "utils.data.error_handling",
    "SquareResizeCrop": "utils.data.transforms",
    "UpperLeftCrop": "utils.data.transforms",
    "denormalize": "utils.data.transforms",
    "make_same_size": "utils.data.transforms",
    "make_size": "utils.data.transforms",
    "pil_to_tensor": "utils.data.transforms",
    "random_square_crop": "utils.data.transforms",
    "random_square_crop_multi": "utils.data.transforms",
    "square_crop": "utils.data.transforms",
    "square_resize_crop": "utils.data.transforms",
    "upper_left_crop": "utils.data.transforms",
    "check_md5": "utils.experiments",
    "check_sha256": "utils.experiments",
    "ensure_downloaded": "utils.experiments",
    "get_git_state": "utils.experiments",
    "get_state_dict_prefix": "utils.experiments",
    "get_training_code": "utils.experiments",
    "save_config": "utils.experiments",
    "test_dataset": "utils.experiments",
    "read_gif": "utils.io",
    "write_gif": "utils.io",
    "ColoredFormatter": "utils.logging",
    "configure_logging": "utils.logging",
    "TextBlock": "utils.text",
    "colored": "utils.text",
    "format_datetime": "utils.text",
    "format_timedelta": "utils.text",
    "outlined": "utils.text",
    "render_text_blocks": "utils.text",
    "show_error": "utils.text",
    "show_warning": "utils.text",
    "uncolored": "utils.text",
    "wrapped": "utils.text",
    "TokenReader": "utils.tokens",
    "TokenWriter": "utils.tokens",
    "token_file": "utils.tokens",
}

# Need to manually set some values which can't be auto-generated.
NAME_MAP.update(
    {
        "ActivationType": "nn.activations",
        "CollateMode": "utils.data.collate",
        "EmbeddingKind": "nn.embeddings",
        "NormType": "nn.norms",
        "ODESolverType": "nn.diffusion.ode",
        "OptType": "task.mixins.optimizer",
        "ParametrizationNormType": "nn.norms",
        "Phase": "core.state",
        "RwkvAttentionState": "nn.architectures.rwkv",
        "RwkvFeedForwardState": "nn.architectures.rwkv",
        "RwkvState": "nn.architectures.rwkv",
    },
)


def __getattr__(name: str) -> object:
    if name not in NAME_MAP:
        raise AttributeError(f"{__name__} has no attribute {name}")

    module_name = f"mlfab.{NAME_MAP[name]}"
    module = __import__(module_name, fromlist=[name])
    return getattr(module, name)


if IMPORT_ALL or TYPE_CHECKING:
    from mlfab.core.conf import (
        UserConfig,
        field,
        get_data_dir,
        get_pretrained_models_dir,
        get_run_dir,
        load_user_config,
    )
    from mlfab.core.state import Phase, State, cast_phase
    from mlfab.nn.activations import (
        ActivationType,
        Clamp,
        Clamp6,
        FastGELU,
        LaplaceActivation,
        QuickGELU,
        ReLUSquared,
        cast_activation_type,
        get_activation,
    )
    from mlfab.nn.architectures.attention import (
        MultiheadAttention,
        NextTokenTransformer,
        NextTokenWithEmbeddingsTransformer,
        TransformerDecoder,
        TransformerDecoderLayer,
        TransformerEncoder,
        TransformerEncoderLayer,
        get_attention_mask,
    )
    from mlfab.nn.architectures.bifpn import BiFPN, BiFPNBlock
    from mlfab.nn.architectures.monotonic_attention import MonotonicAttention, monotonic_attention
    from mlfab.nn.architectures.next_token import sample_from_logits, top_k_sampling, top_p_sampling
    from mlfab.nn.architectures.rwkv import (
        NextTokenRwkv,
        NextTokenWithEmbeddingsRwkv,
        RwkvAttention,
        RwkvAttentionState,
        RwkvBlock,
        RwkvFeedForward,
        RwkvFeedForwardState,
        RwkvStack,
        RwkvState,
    )
    from mlfab.nn.architectures.unet import UNet
    from mlfab.nn.device.auto import detect_device
    from mlfab.nn.device.base import base_device
    from mlfab.nn.device.cpu import cpu_device
    from mlfab.nn.device.gpu import gpu_device
    from mlfab.nn.device.metal import metal_device
    from mlfab.nn.diffusion.consistency import ConsistencyModel
    from mlfab.nn.diffusion.gaussian import GaussianDiffusion, cast_beta_schedule, get_diffusion_beta_schedule
    from mlfab.nn.diffusion.ode import (
        BaseODESolver,
        EulerODESolver,
        HeunODESolver,
        ODESolverType,
        RK4ODESolver,
        cast_solver_type,
        get_ode_solver,
    )
    from mlfab.nn.embeddings import (
        EmbeddingKind,
        FourierEmbeddings,
        IdentityPositionalEmbeddings,
        LearnedPositionalEmbeddings,
        RotaryEmbeddings,
        SinusoidalEmbeddings,
        apply_rotary_embeddings,
        cast_embedding_kind,
        fourier_embeddings,
        get_positional_embeddings,
        get_rotary_embeddings,
        rotary_embeddings,
    )
    from mlfab.nn.functions import (
        DropPath,
        StreamingConv1d,
        StreamingConvTranspose1d,
        as_numpy_array,
        as_tensor,
        device_fn,
        drop_path,
        gated_residual,
        recursive_apply,
        recursive_chunk,
        residual,
        set_random_seed,
        streamable_cbr,
        streaming_add,
        streaming_conv_1d,
        streaming_conv_transpose_1d,
    )
    from mlfab.nn.init import init_
    from mlfab.nn.kmeans import KMeans, kmeans_fn
    from mlfab.nn.lora import (
        LoraColumnParallelLinear,
        LoraConv1d,
        LoraConv2d,
        LoraConvTranspose1d,
        LoraConvTranspose2d,
        LoraEmbedding,
        LoraGRU,
        LoraGRUCell,
        LoraLinear,
        LoraLSTM,
        LoraLSTMCell,
        LoraParallelEmbedding,
        LoraRowParallelLinear,
        freeze_non_lora_,
        lora,
        maybe_lora,
        maybe_lora_weight_norm,
        reset_lora_weights_,
    )
    from mlfab.nn.losses import (
        LPIPS,
        ImageGradLoss,
        MelLoss,
        MFCCLoss,
        MultiResolutionSTFTLoss,
        SSIMLoss,
        STFTLoss,
        get_stft_window,
        kl_div_pair_loss,
        kl_div_single_loss,
        log_cosh_loss,
        log_stft_magnitude_loss,
        pseudo_huber_loss,
        spectral_convergence_loss,
        stft,
        stft_magnitude_loss,
    )
    from mlfab.nn.lr_schedulers import (
        BaseLRScheduler,
        ConstantLRScheduler,
        CosineDecayLRScheduler,
        CosineLRScheduler,
        LinearLRScheduler,
        SchedulerAdapter,
    )
    from mlfab.nn.norms import (
        ConvLayerNorm,
        LastBatchNorm,
        NormType,
        ParametrizationNormType,
        cast_norm_type,
        cast_parametrize_norm_type,
        get_norm_1d,
        get_norm_2d,
        get_norm_3d,
        get_norm_linear,
        get_parametrization_norm,
    )
    from mlfab.nn.optimizers import Adam, Lion, can_use_foreach, can_use_fused, separate_decayable_params
    from mlfab.nn.parallel import (
        ColumnParallelLinear,
        MultiProcessConfig,
        ParallelConfig,
        ParallelEmbedding,
        RowParallelLinear,
        ddp,
        dp,
        fsdp,
        get_data_worker_info,
        get_local_rank,
        get_local_rank_optional,
        get_local_world_size,
        get_local_world_size_optional,
        get_master_addr,
        get_master_port,
        get_rank,
        get_rank_optional,
        get_unused_port,
        get_world_size,
        get_world_size_optional,
        is_distributed,
        is_master,
        launch_subprocesses,
        split_n_items_across_workers,
    )
    from mlfab.nn.quantization.fsq import FiniteScalarQuantization
    from mlfab.nn.quantization.lfq import LookupFreeQuantization
    from mlfab.nn.quantization.vq import ResidualVectorQuantization, VectorQuantization
    from mlfab.task.launchers.base import BaseLauncher
    from mlfab.task.launchers.cli import CliLauncher
    from mlfab.task.launchers.multi_process import MultiProcessLauncher
    from mlfab.task.launchers.single_process import SingleProcessLauncher
    from mlfab.task.launchers.slurm import SlurmLauncher
    from mlfab.task.logger import LogAudio, Logger, LoggerImpl, LogImage, LogLine, LogVideo, make_square_image_or_video
    from mlfab.task.loggers.json import JsonLogger
    from mlfab.task.loggers.state import StateLogger
    from mlfab.task.loggers.stdout import StdoutLogger
    from mlfab.task.loggers.tensorboard import TensorboardLogger
    from mlfab.task.mixins.compile import TorchCompileOptions
    from mlfab.task.mixins.cpu_stats import CPUStatsOptions
    from mlfab.task.mixins.data_loader import DataloaderConfig
    from mlfab.task.mixins.gpu_stats import GPUStatsOptions
    from mlfab.task.mixins.optimizer import OptType
    from mlfab.task.mixins.profiler import ProfilerOptions
    from mlfab.task.script import Script, ScriptConfig
    from mlfab.task.task import Config, Task
    from mlfab.utils.data.collate import CollateMode, collate, collate_nullable, pad_all, pad_sequence
    from mlfab.utils.data.dataset import SmallDataset
    from mlfab.utils.data.error_handling import (
        ErrorHandlingDataset,
        ErrorHandlingIterableDataset,
        ExceptionSummary,
        error_handling_dataset,
    )
    from mlfab.utils.data.transforms import (
        SquareResizeCrop,
        UpperLeftCrop,
        denormalize,
        make_same_size,
        make_size,
        pil_to_tensor,
        random_square_crop,
        random_square_crop_multi,
        square_crop,
        square_resize_crop,
        upper_left_crop,
    )
    from mlfab.utils.experiments import (
        check_md5,
        check_sha256,
        ensure_downloaded,
        get_git_state,
        get_state_dict_prefix,
        get_training_code,
        save_config,
        test_dataset,
    )
    from mlfab.utils.io import read_gif, write_gif
    from mlfab.utils.logging import ColoredFormatter, configure_logging
    from mlfab.utils.text import (
        TextBlock,
        colored,
        format_datetime,
        format_timedelta,
        outlined,
        render_text_blocks,
        show_error,
        show_warning,
        uncolored,
        wrapped,
    )
    from mlfab.utils.tokens import TokenReader, TokenWriter, token_file

del TYPE_CHECKING, IMPORT_ALL



"""Defines a mixin for abstracting the PyTorch tensor device."""

import functools
from dataclasses import dataclass
from typing import Generic, TypeVar

import torch

from mlfab.core.conf import Device as BaseDeviceConfig, field, parse_dtype
from mlfab.nn.device.auto import detect_device
from mlfab.nn.device.base import base_device
from mlfab.task.base import BaseConfig, BaseTask


@dataclass
class DeviceConfig(BaseConfig):
    device: BaseDeviceConfig = field(BaseDeviceConfig(), help="Device configuration")


Config = TypeVar("Config", bound=DeviceConfig)


class DeviceMixin(BaseTask[Config], Generic[Config]):
    @functools.cached_property
    def device(self) -> base_device:
        return detect_device()

    @functools.cached_property
    def torch_device(self) -> torch.device:
        return self.device.device

    @functools.cached_property
    def torch_dtype(self) -> torch.dtype:
        if (dtype := parse_dtype(self.config.device)) is not None:
            return dtype
        return self.device.dtype



"""Defines a mixin for instantiating dataloaders."""

import logging
from dataclasses import dataclass
from typing import Generic, TypeVar

from dpshdl.dataloader import Dataloader
from dpshdl.dataset import Dataset, ErrorHandlingDataset
from omegaconf import II, MISSING

from mlfab.core.conf import field, is_missing, load_user_config
from mlfab.core.state import Phase
from mlfab.nn.functions import set_random_seed
from mlfab.nn.parallel import get_rank
from mlfab.task.base import BaseConfig, BaseTask
from mlfab.task.mixins.process import ProcessConfig, ProcessMixin

logger = logging.getLogger(__name__)

Sample = TypeVar("Sample")
Batch = TypeVar("Batch")

T = TypeVar("T")
Tc = TypeVar("Tc")


@dataclass
class DataloaderConfig:
    num_workers: int = field(MISSING, help="Number of workers for loading samples")
    prefetch_factor: int = field(2, help="Number of items to pre-fetch on the host")


@dataclass
class DataloadersConfig(ProcessConfig, BaseConfig):
    batch_size: int = field(MISSING, help="Size of each batch")
    train_dl: DataloaderConfig = field(
        DataloaderConfig(num_workers=II("mlfab.num_workers:-1")),
        help="Train dataloader config",
    )
    test_dl: DataloaderConfig = field(
        DataloaderConfig(num_workers=1),
        help="Valid dataloader config",
    )
    debug_dataloader: bool = field(False, help="Debug dataloaders")
    use_pytorch_dataloader: bool = field(False, help="Use PyTorch dataloaders")


Config = TypeVar("Config", bound=DataloadersConfig)


class DataloadersMixin(ProcessMixin[Config], BaseTask[Config], Generic[Config]):
    def __init__(self, config: Config) -> None:
        if is_missing(config, "batch_size"):
            config.batch_size = self.get_batch_size()

        super().__init__(config)

    def get_batch_size(self) -> int:
        raise NotImplementedError(
            "When `batch_size` is not specified in your training config, you should override the `get_batch_size` "
            "method to return the desired training batch size."
        )

    def dataloader_config(self, phase: Phase) -> DataloaderConfig:
        match phase:
            case "train":
                return self.config.train_dl
            case "valid":
                return self.config.test_dl
            case "test":
                return self.config.test_dl
            case _:
                raise KeyError(f"Unknown phase: {phase}")

    def get_dataset(self, phase: Phase) -> Dataset:
        """Returns the dataset for the given phase.

        Args:
            phase: The phase for the dataset to return.

        Raises:
            NotImplementedError: If this method is not overridden
        """
        raise NotImplementedError("The task should implement `get_dataset`")

    def get_dataloader(
        self,
        dataset: Dataset[Sample, Batch],
        phase: Phase,
    ) -> Dataloader[Sample, Batch]:
        debugging = self.config.debug_dataloader
        if debugging:
            logger.warning("Parallel dataloaders disabled in debugging mode")

        conf = load_user_config()
        if conf.error_handling.enabled:
            dataset = ErrorHandlingDataset(
                dataset,
                sleep_backoff=conf.error_handling.sleep_backoff,
                sleep_backoff_power=conf.error_handling.sleep_backoff_power,
                maximum_exceptions=conf.error_handling.maximum_exceptions,
                backoff_after=conf.error_handling.backoff_after,
                traceback_depth=conf.error_handling.exception_location_traceback_depth,
                flush_every_n_seconds=conf.error_handling.flush_exception_summary_every,
                flush_every_n_steps=conf.error_handling.flush_exception_summary_every,
            )

        cfg = self.dataloader_config(phase)

        # Creates a globally unique name for each dataloader.
        rank = get_rank()
        name = f"{phase}-{rank}"

        return Dataloader(
            dataset=dataset,
            num_workers=0 if debugging else cfg.num_workers,
            batch_size=self.config.batch_size,
            prefetch_factor=cfg.prefetch_factor,
            mp_context=None if debugging or cfg.num_workers < 1 else self.multiprocessing_context,
            mp_manager=None if debugging or cfg.num_workers < 1 else self.multiprocessing_manager,
            collate_worker_init_fn=self.collate_worker_init_fn,
            dataloader_worker_init_fn=self.data_worker_init_fn,
            name=name,
        )

    @classmethod
    def data_worker_init_fn(cls, worker_id: int, num_workers: int) -> None:
        set_random_seed(offset=worker_id + 1)

    @classmethod
    def collate_worker_init_fn(cls) -> None:
        set_random_seed(offset=0)
